
/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/

#ifndef _ASM_MACROS_H
#define _ASM_MACROS_H

#include "types.h"
#include "mipsregs.h"
#include "irq.h"

#ifndef __ASSEMBLY__

/***************************************************************************/
/* xlr_read_32bit_cp0_index() - Read index (32 bit) from co-processor 0        */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MF     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00000  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint32_t xlr_read_32bit_cp0_index(void)
{
  uint32_t val;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x40010000  \n\t"  /* mfc0 at, $0 (index), 0 (sel) */
      "nop               \n\t"
      "move   %0, $1     \n\t"

      ".set pop\n"

      : "=r" (val));

  return val;
}

/***************************************************************************/
/* xlr_read_32bit_cp0_random() - Read random (32 bit) from co-processor 0        */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MF     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00000  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint32_t xlr_read_32bit_cp0_random(void)
{
  uint32_t val;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x40010800  \n\t"  /* mfc0 at, $1 (random), 0 (sel) */
      "nop               \n\t"
      "move   %0, $1     \n\t"

      ".set pop\n"

      : "=r" (val));

  return val;
}

/***************************************************************************/
/* xlr_read_64bit_cp0_entrylo0() - Read entrylo0 (64 bit) from co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMF    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00001  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint64_t xlr_read_64bit_cp0_entrylo0(void)
{
  uint32_t high, low;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x40211000  \n\t"  /* dmfc0 at, $2 (entrylo0), 0 (sel) */
      "nop               \n\t"
      "dsra32 %0, $1, 0  \n\t"
      "sll    %1, $1, 0  \n\t"

      ".set pop\n"

      : "=r" (high), "=r" (low)
  );

  return ( ((uint64_t)high) << 32) | low;
}

/***************************************************************************/
/* xlr_read_32bit_cp0_conf() - Read conf (32 bit) from co-processor 0        */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MF     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00000  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint32_t xlr_read_32bit_cp0_conf(void)
{
  uint32_t val;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x40011800  \n\t"  /* mfc0 at, $3 (entrylo1), 0 (sel) */
      "nop               \n\t"
      "move   %0, $1     \n\t"

      ".set pop\n"

      : "=r" (val));

  return val;
}

/***************************************************************************/
/* xlr_read_64bit_cp0_entrylo1() - Read entrylo1 (64 bit) from co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMF    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00001  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint64_t xlr_read_64bit_cp0_entrylo1(void)
{
  uint32_t high, low;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x40211800  \n\t"  /* dmfc0 at, $3 (entrylo1), 0 (sel) */
      "nop               \n\t"
      "dsra32 %0, $1, 0  \n\t"
      "sll    %1, $1, 0  \n\t"

      ".set pop\n"

      : "=r" (high), "=r" (low)
  );

  return ( ((uint64_t)high) << 32) | low;
}

/***************************************************************************/
/* xlr_read_64bit_cp0_context() - Read context (64 bit) from co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMF    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00001  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint64_t xlr_read_64bit_cp0_context(void)
{
  uint32_t high, low;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x40212000  \n\t"  /* dmfc0 at, $4 (context), 0 (sel) */
      "nop               \n\t"
      "dsra32 %0, $1, 0  \n\t"
      "sll    %1, $1, 0  \n\t"

      ".set pop\n"

      : "=r" (high), "=r" (low)
  );

  return ( ((uint64_t)high) << 32) | low;
}

/***************************************************************************/
/* xlr_read_32bit_cp0_pagemask() - Read pagemask (32 bit) from co-processor 0        */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MF     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00000  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint32_t xlr_read_32bit_cp0_pagemask(void)
{
  uint32_t val;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x40012800  \n\t"  /* mfc0 at, $5 (pagemask), 0 (sel) */
      "nop               \n\t"
      "move   %0, $1     \n\t"

      ".set pop\n"

      : "=r" (val));

  return val;
}

/***************************************************************************/
/* xlr_read_32bit_cp0_wired() - Read wired (32 bit) from co-processor 0        */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MF     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00000  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint32_t xlr_read_32bit_cp0_wired(void)
{
  uint32_t val;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x40010006  \n\t"  /* mfc0 at, $6 (wired), 0 (sel) */
      "nop               \n\t"
      "move   %0, $1     \n\t"

      ".set pop\n"

      : "=r" (val));

  return val;
}

/***************************************************************************/
/* xlr_read_32bit_cp0_reserved0() - Read reserved0 (32 bit) from co-processor 0        */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MF     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00000  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint32_t xlr_read_32bit_cp0_reserved0(void)
{
  uint32_t val;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x40013800  \n\t"  /* mfc0 at, $7 (reserved), 0 (sel) */
      "nop               \n\t"
      "move   %0, $1     \n\t"

      ".set pop\n"

      : "=r" (val));

  return val;
}

/***************************************************************************/
/* xlr_read_64bit_cp0_badvaddr() - Read badvaddr (64 bit) from co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMF    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00001  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint64_t xlr_read_64bit_cp0_badvaddr(void)
{
  uint32_t high, low;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x40214000  \n\t"  /* dmfc0 at, $8 (badvaddr), 0 (sel) */
      "nop               \n\t"
      "dsra32 %0, $1, 0  \n\t"
      "sll    %1, $1, 0  \n\t"

      ".set pop\n"

      : "=r" (high), "=r" (low)
  );

  return ( ((uint64_t)high) << 32) | low;
}

/***************************************************************************/
/* xlr_read_32bit_cp0_count() - Read count (32 bit) from co-processor 0        */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MF     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00000  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint32_t xlr_read_32bit_cp0_count(void)
{
  uint32_t val;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x40014800  \n\t"  /* mfc0 at, $9 (count), 0 (sel) */
      "nop               \n\t"
      "move   %0, $1     \n\t"

      ".set pop\n"

      : "=r" (val));

  return val;
}

/***************************************************************************/
/* xlr_read_64bit_cp0_eirr() - Read eirr (64 bit) from co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMF    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00001  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint64_t xlr_read_64bit_cp0_eirr(void)
{
  uint32_t high, low;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x40214806  \n\t"  /* dmfc0 at, $9 (eirr), 6 (sel) */
      "nop               \n\t"
      "dsra32 %0, $1, 0  \n\t"
      "sll    %1, $1, 0  \n\t"

      ".set pop\n"

      : "=r" (high), "=r" (low)
  );

  return ( ((uint64_t)high) << 32) | low;
}

/***************************************************************************/
/* xlr_read_64bit_cp0_eimr() - Read eimr (64 bit) from co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMF    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00001  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint64_t xlr_read_64bit_cp0_eimr(void)
{
  uint32_t high, low;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x40214807  \n\t"  /* dmfc0 at, $9 (eimr), 7 (sel) */
      "nop               \n\t"
      "dsra32 %0, $1, 0  \n\t"
      "sll    %1, $1, 0  \n\t"

      ".set pop\n"

      : "=r" (high), "=r" (low)
  );

  return ( ((uint64_t)high) << 32) | low;
}

/***************************************************************************/
/* xlr_read_64bit_cp0_entryhi() - Read entryhi (64 bit) from co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMF    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00001  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint64_t xlr_read_64bit_cp0_entryhi(void)
{
  uint32_t high, low;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x40215000  \n\t"  /* dmfc0 at, $10 (entryhi), 0 (sel) */
      "nop               \n\t"
      "dsra32 %0, $1, 0  \n\t"
      "sll    %1, $1, 0  \n\t"

      ".set pop\n"

      : "=r" (high), "=r" (low)
  );

  return ( ((uint64_t)high) << 32) | low;
}

/***************************************************************************/
/* xlr_read_32bit_cp0_compare() - Read compare (32 bit) from co-processor 0        */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MF     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00000  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint32_t xlr_read_32bit_cp0_compare(void)
{
  uint32_t val;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x40015800  \n\t"  /* mfc0 at, $11 (compare), 0 (sel) */
      "nop               \n\t"
      "move   %0, $1     \n\t"

      ".set pop\n"

      : "=r" (val));

  return val;
}

/***************************************************************************/
/* xlr_read_32bit_cp0_status() - Read status (32 bit) from co-processor 0        */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MF     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00000  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint32_t xlr_read_32bit_cp0_status(void)
{
  uint32_t val;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x40016000  \n\t"  /* mfc0 at, $12 (status), 0 (sel) */
      "nop               \n\t"
      "move   %0, $1     \n\t"

      ".set pop\n"

      : "=r" (val));

  return val;
}

/***************************************************************************/
/* xlr_read_32bit_cp0_cause() - Read cause (32 bit) from co-processor 0        */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MF     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00000  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint32_t xlr_read_32bit_cp0_cause(void)
{
  uint32_t val;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x40016800  \n\t"  /* mfc0 at, $13 (cause), 0 (sel) */
      "nop               \n\t"
      "move   %0, $1     \n\t"

      ".set pop\n"

      : "=r" (val));

  return val;
}

/***************************************************************************/
/* xlr_read_64bit_cp0_epc() - Read epc (64 bit) from co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMF    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00001  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint64_t xlr_read_64bit_cp0_epc(void)
{
  uint32_t high, low;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x40217000  \n\t"  /* dmfc0 at, $14 (epc), 0 (sel) */
      "nop               \n\t"
      "dsra32 %0, $1, 0  \n\t"
      "sll    %1, $1, 0  \n\t"

      ".set pop\n"

      : "=r" (high), "=r" (low)
  );

  return ( ((uint64_t)high) << 32) | low;
}

/***************************************************************************/
/* xlr_read_32bit_cp0_processor_id() - Read processor_id (32 bit) from co-processor 0        */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MF     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00000  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint32_t xlr_read_32bit_cp0_processor_id(void)
{
  uint32_t val;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x40017800  \n\t"  /* mfc0 at, $15 (processor_id), 0 (sel) */
      "nop               \n\t"
      "move   %0, $1     \n\t"

      ".set pop\n"

      : "=r" (val));

  return val;
}

/***************************************************************************/
/* xlr_read_32bit_cp0_config0() - Read config0 (32 bit) from co-processor 0        */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MF     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00000  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint32_t xlr_read_32bit_cp0_config0(void)
{
  uint32_t val;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x40018000  \n\t"  /* mfc0 at, $16 (config0), 0 (sel) */
      "nop               \n\t"
      "move   %0, $1     \n\t"

      ".set pop\n"

      : "=r" (val));

  return val;
}

/***************************************************************************/
/* xlr_read_32bit_cp0_config1() - Read config1 (32 bit) from co-processor 0        */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MF     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00000  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint32_t xlr_read_32bit_cp0_config1(void)
{
  uint32_t val;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x40018001  \n\t"  /* mfc0 at, $16 (config1), 1 (sel) */
      "nop               \n\t"
      "move   %0, $1     \n\t"

      ".set pop\n"

      : "=r" (val));

  return val;
}

/***************************************************************************/
/* xlr_read_64bit_cp0_lladdr() - Read lladdr (64 bit) from co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMF    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00001  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint64_t xlr_read_64bit_cp0_lladdr(void)
{
  uint32_t high, low;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x40218800  \n\t"  /* dmfc0 at, $17 (reserved), 0 (sel) */
      "nop               \n\t"
      "dsra32 %0, $1, 0  \n\t"
      "sll    %1, $1, 0  \n\t"

      ".set pop\n"

      : "=r" (high), "=r" (low)
  );

  return ( ((uint64_t)high) << 32) | low;
}

/***************************************************************************/
/* xlr_read_64bit_cp0_watchlo() - Read watchlo (64 bit) from co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMF    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00001  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint64_t xlr_read_64bit_cp0_watchlo(void)
{
  uint32_t high, low;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x40219000  \n\t"  /* dmfc0 at, $18 (watchlo), 0 (sel) */
      "nop               \n\t"
      "dsra32 %0, $1, 0  \n\t"
      "sll    %1, $1, 0  \n\t"

      ".set pop\n"

      : "=r" (high), "=r" (low)
  );

  return ( ((uint64_t)high) << 32) | low;
}

/***************************************************************************/
/* xlr_read_32bit_cp0_watchhi() - Read watchhi (32 bit) from co-processor 0        */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MF     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00000  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint32_t xlr_read_32bit_cp0_watchhi(void)
{
  uint32_t val;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x40019800  \n\t"  /* mfc0 at, $19 (watchhi), 0 (sel) */
      "nop               \n\t"
      "move   %0, $1     \n\t"

      ".set pop\n"

      : "=r" (val));

  return val;
}

/***************************************************************************/
/* xlr_read_64bit_cp0_xcontext() - Read xcontext (64 bit) from co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMF    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00001  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint64_t xlr_read_64bit_cp0_xcontext(void)
{
  uint32_t high, low;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x4021a000  \n\t"  /* dmfc0 at, $20 (xcontext), 0 (sel) */
      "nop               \n\t"
      "dsra32 %0, $1, 0  \n\t"
      "sll    %1, $1, 0  \n\t"

      ".set pop\n"

      : "=r" (high), "=r" (low)
  );

  return ( ((uint64_t)high) << 32) | low;
}

/***************************************************************************/
/* xlr_read_32bit_cp0_reserved2() - Read reserved2 (32 bit) from co-processor 0        */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MF     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00000  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint32_t xlr_read_32bit_cp0_reserved2(void)
{
  uint32_t val;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x4001a800  \n\t"  /* mfc0 at, $21 (reserved), 0 (sel) */
      "nop               \n\t"
      "move   %0, $1     \n\t"

      ".set pop\n"

      : "=r" (val));

  return val;
}

/***************************************************************************/
/* xlr_read_32bit_cp0_diagnostic() - Read diagnostic (32 bit) from co-processor 0        */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MF     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00000  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint32_t xlr_read_32bit_cp0_diagnostic(void)
{
  uint32_t val;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x4001b000  \n\t"  /* mfc0 at, $22 (scratch0), 0 (sel) */
      "nop               \n\t"
      "move   %0, $1     \n\t"

      ".set pop\n"

      : "=r" (val));

  return val;
}

/***************************************************************************/
/* xlr_read_64bit_cp0_scratch0() - Read scratch0 (64 bit) from co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMF    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00001  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint64_t xlr_read_64bit_cp0_scratch0(void)
{
  uint32_t high, low;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x4021b000  \n\t"  /* dmfc0 at, $22 (scratch0), 0 (sel) */
      "nop               \n\t"
      "dsra32 %0, $1, 0  \n\t"
      "sll    %1, $1, 0  \n\t"

      ".set pop\n"

      : "=r" (high), "=r" (low)
  );

  return ( ((uint64_t)high) << 32) | low;
}

/***************************************************************************/
/* xlr_read_64bit_cp0_scratch1() - Read scratch1 (64 bit) from co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMF    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00001  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint64_t xlr_read_64bit_cp0_scratch1(void)
{
  uint32_t high, low;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x4021b001  \n\t"  /* dmfc0 at, $22 (scratch1), 1 (sel) */
      "nop               \n\t"
      "dsra32 %0, $1, 0  \n\t"
      "sll    %1, $1, 0  \n\t"

      ".set pop\n"

      : "=r" (high), "=r" (low)
  );

  return ( ((uint64_t)high) << 32) | low;
}

/***************************************************************************/
/* xlr_read_64bit_cp0_scratch2() - Read scratch2 (64 bit) from co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMF    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00001  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint64_t xlr_read_64bit_cp0_scratch2(void)
{
  uint32_t high, low;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x4021b002  \n\t"  /* dmfc0 at, $22 (scratch2), 2 (sel) */
      "nop               \n\t"
      "dsra32 %0, $1, 0  \n\t"
      "sll    %1, $1, 0  \n\t"

      ".set pop\n"

      : "=r" (high), "=r" (low)
  );

  return ( ((uint64_t)high) << 32) | low;
}

/***************************************************************************/
/* xlr_read_64bit_cp0_scratch3() - Read scratch3 (64 bit) from co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMF    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00001  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint64_t xlr_read_64bit_cp0_scratch3(void)
{
  uint32_t high, low;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x4021b003  \n\t"  /* dmfc0 at, $22 (scratch3), 3 (sel) */
      "nop               \n\t"
      "dsra32 %0, $1, 0  \n\t"
      "sll    %1, $1, 0  \n\t"

      ".set pop\n"

      : "=r" (high), "=r" (low)
  );

  return ( ((uint64_t)high) << 32) | low;
}

/***************************************************************************/
/* xlr_read_64bit_cp0_scratch4() - Read scratch4 (64 bit) from co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMF    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00001  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint64_t xlr_read_64bit_cp0_scratch4(void)
{
  uint32_t high, low;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x4021b004  \n\t"  /* dmfc0 at, $22 (scratch4), 4 (sel) */
      "nop               \n\t"
      "dsra32 %0, $1, 0  \n\t"
      "sll    %1, $1, 0  \n\t"

      ".set pop\n"

      : "=r" (high), "=r" (low)
  );

  return ( ((uint64_t)high) << 32) | low;
}

/***************************************************************************/
/* xlr_read_64bit_cp0_scratch5() - Read scratch5 (64 bit) from co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMF    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00001  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint64_t xlr_read_64bit_cp0_scratch5(void)
{
  uint32_t high, low;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x4021b005  \n\t"  /* dmfc0 at, $22 (scratch5), 5 (sel) */
      "nop               \n\t"
      "dsra32 %0, $1, 0  \n\t"
      "sll    %1, $1, 0  \n\t"

      ".set pop\n"

      : "=r" (high), "=r" (low)
  );

  return ( ((uint64_t)high) << 32) | low;
}

/***************************************************************************/
/* xlr_read_64bit_cp0_scratch6() - Read scratch6 (64 bit) from co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMF    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00001  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint64_t xlr_read_64bit_cp0_scratch6(void)
{
  uint32_t high, low;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x4021b006  \n\t"  /* dmfc0 at, $22 (scratch6), 6 (sel) */
      "nop               \n\t"
      "dsra32 %0, $1, 0  \n\t"
      "sll    %1, $1, 0  \n\t"

      ".set pop\n"

      : "=r" (high), "=r" (low)
  );

  return ( ((uint64_t)high) << 32) | low;
}

/***************************************************************************/
/* xlr_read_64bit_cp0_scratch7() - Read scratch7 (64 bit) from co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMF    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00001  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint64_t xlr_read_64bit_cp0_scratch7(void)
{
  uint32_t high, low;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x4021b007  \n\t"  /* dmfc0 at, $22 (scratch7), 7 (sel) */
      "nop               \n\t"
      "dsra32 %0, $1, 0  \n\t"
      "sll    %1, $1, 0  \n\t"

      ".set pop\n"

      : "=r" (high), "=r" (low)
  );

  return ( ((uint64_t)high) << 32) | low;
}

/***************************************************************************/
/* xlr_read_32bit_cp0_debug() - Read debug (32 bit) from co-processor 0        */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MF     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00000  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint32_t xlr_read_32bit_cp0_debug(void)
{
  uint32_t val;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x4001b800  \n\t"  /* mfc0 at, $23 (debug), 0 (sel) */
      "nop               \n\t"
      "move   %0, $1     \n\t"

      ".set pop\n"

      : "=r" (val));

  return val;
}

/***************************************************************************/
/* xlr_read_32bit_cp0_depc() - Read depc (32 bit) from co-processor 0        */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MF     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00000  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint32_t xlr_read_32bit_cp0_depc(void)
{
  uint32_t val;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x4001c000  \n\t"  /* mfc0 at, $24 (depc), 0 (sel) */
      "nop               \n\t"
      "move   %0, $1     \n\t"

      ".set pop\n"

      : "=r" (val));

  return val;
}

/***************************************************************************/
/* xlr_read_32bit_cp0_perfctr0_low() - Read perfctr0_low (32 bit) from co-processor 0        */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MF     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00000  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint32_t xlr_read_32bit_cp0_perfctr0_low(void)
{
  uint32_t val;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x4001c800  \n\t"  /* mfc0 at, $25 (perfctr0), 0 (sel) */
      "nop               \n\t"
      "move   %0, $1     \n\t"

      ".set pop\n"

      : "=r" (val));

  return val;
}

/***************************************************************************/
/* xlr_read_32bit_cp0_perfctr0_hi() - Read perfctr0_hi (32 bit) from co-processor 0        */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MF     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00000  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint32_t xlr_read_32bit_cp0_perfctr0_hi(void)
{
  uint32_t val;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x4001c801  \n\t"  /* mfc0 at, $25 (perfctr0), 1 (sel) */
      "nop               \n\t"
      "move   %0, $1     \n\t"

      ".set pop\n"

      : "=r" (val));

  return val;
}

/***************************************************************************/
/* xlr_read_32bit_cp0_perfctr1_low() - Read perfctr1_low (32 bit) from co-processor 0        */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MF     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00000  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint32_t xlr_read_32bit_cp0_perfctr1_low(void)
{
  uint32_t val;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x4001c802  \n\t"  /* mfc0 at, $25 (perfctr1), 2 (sel) */
      "nop               \n\t"
      "move   %0, $1     \n\t"

      ".set pop\n"

      : "=r" (val));

  return val;
}

/***************************************************************************/
/* xlr_read_32bit_cp0_perfctr1_high() - Read perfctr1_high (32 bit) from co-processor 0        */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MF     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00000  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint32_t xlr_read_32bit_cp0_perfctr1_high(void)
{
  uint32_t val;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x4001c803  \n\t"  /* mfc0 at, $25 (perfctr1), 3 (sel) */
      "nop               \n\t"
      "move   %0, $1     \n\t"

      ".set pop\n"

      : "=r" (val));

  return val;
}

/***************************************************************************/
/* xlr_read_32bit_cp0_errorctl() - Read errorctl (32 bit) from co-processor 0        */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MF     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00000  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint32_t xlr_read_32bit_cp0_errorctl(void)
{
  uint32_t val;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x4001d000  \n\t"  /* mfc0 at, $26 (errorctl), 0 (sel) */
      "nop               \n\t"
      "move   %0, $1     \n\t"

      ".set pop\n"

      : "=r" (val));

  return val;
}

/***************************************************************************/
/* xlr_read_32bit_cp0_cacheerr() - Read cacheerr (32 bit) from co-processor 0        */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MF     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00000  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint32_t xlr_read_32bit_cp0_cacheerr(void)
{
  uint32_t val;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x4001d800  \n\t"  /* mfc0 at, $27 (cacheerr), 0 (sel) */
      "nop               \n\t"
      "move   %0, $1     \n\t"

      ".set pop\n"

      : "=r" (val));

  return val;
}


/***************************************************************************/
/* xlr_read_64bit_cp0_taglo0() - Read taglo0 (64 bit) from co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMF    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00001  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint64_t xlr_read_64bit_cp0_taglo0(void)
{
  uint32_t high, low;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x4021e000  \n\t"  /* dmfc0 at, $28 (taglo), 0 (sel) */
      "nop               \n\t"
      "dsra32 %0, $1, 0  \n\t"
      "sll    %1, $1, 0  \n\t"

      ".set pop\n"

      : "=r" (high), "=r" (low)
  );

  return ( ((uint64_t)high) << 32) | low;
}

/***************************************************************************/
/* xlr_read_64bit_cp0_taglo1() - Read taglo1 (64 bit) from co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMF    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00001  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint64_t xlr_read_64bit_cp0_taglo1(void)
{
  uint32_t high, low;

   __asm__ __volatile__ (
    ".set push\n"
    ".set noreorder\n"
    ".set noat\n"
    ".set mips4\n"

    ".word 0x4021e001  \n\t"  /* dmfc0 at, $28 (taglo), 1 (sel) */
    "nop               \n\t"
    "dsra32 %0, $1, 0  \n\t"
    "sll    %1, $1, 0  \n\t"

     

    ".set pop\n"

     

    : "=r" (high), "=r" (low)
  );
  return ( ((uint64_t)high) << 32) | low;
}


/***************************************************************************/
/* xlr_read_64bit_cp0_taglo2() - Read taglo2 (64 bit) from co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMF    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00001  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint64_t xlr_read_64bit_cp0_taglo2(void)
{
  uint32_t high, low;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x4021e002  \n\t"  /* dmfc0 at, $28 (taglo), 2 (sel) */
      "nop               \n\t"
      "dsra32 %0, $1, 0  \n\t"
      "sll    %1, $1, 0  \n\t"

      ".set pop\n"

      : "=r" (high), "=r" (low)
  );

  return ( ((uint64_t)high) << 32) | low;
}


/***************************************************************************/
/* xlr_read_64bit_cp0_taghi0() - Read taghi0 (64 bit) from co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMF    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00001  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint64_t xlr_read_64bit_cp0_taghi0(void)
{
  uint32_t high, low;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x4021e800  \n\t"  /* dmfc0 at, $29 (taghi), 0 (sel) */
      "nop               \n\t"
      "dsra32 %0, $1, 0  \n\t"
      "sll    %1, $1, 0  \n\t"

      ".set pop\n"

      : "=r" (high), "=r" (low)
  );

  return ( ((uint64_t)high) << 32) | low;
}
/***************************************************************************/
/* xlr_read_64bit_cp0_taghi1() - Read taghi1 (64 bit) from co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMF    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00001  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint64_t xlr_read_64bit_cp0_taghi1(void)
{
   uint32_t high, low;

   __asm__ __volatile__ (

      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"


      ".word 0x4021e801  \n\t"  /* dmfc0 at, $29 (taghi), 1 (sel) */
      "nop               \n\t"

      "dsra32 %0, $1, 0  \n\t"
      "sll    %1, $1, 0  \n\t"


      ".set pop\n"

      
      : "=r" (high), "=r" (low)

   );
   return ( ((uint64_t)high) << 32) | low;
}


/***************************************************************************/
/* xlr_read_64bit_cp0_taghi2() - Read taghi2 (64 bit) from co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMF    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00001  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint64_t xlr_read_64bit_cp0_taghi2(void)
{
  uint32_t high, low;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x4021e802  \n\t"  /* dmfc0 at, $29 (taghi), 2 (sel) */
      "nop               \n\t"
      "dsra32 %0, $1, 0  \n\t"
      "sll    %1, $1, 0  \n\t"

      ".set pop\n"

      : "=r" (high), "=r" (low)
  );

  return ( ((uint64_t)high) << 32) | low;
}

/* xlr_read_64bit_cp0_errorepc() - Read errorepc (64 bit) from co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMF    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00001  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint64_t xlr_read_64bit_cp0_errorepc(void)
{
  uint32_t high, low;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x4021f000  \n\t"  /* dmfc0 at, $30 (errorepc), 0 (sel) */
      "nop               \n\t"
      "dsra32 %0, $1, 0  \n\t"
      "sll    %1, $1, 0  \n\t"

      ".set pop\n"

      : "=r" (high), "=r" (low)
  );

  return ( ((uint64_t)high) << 32) | low;
}

/***************************************************************************/
/* xlr_read_32bit_cp0_desave() - Read desave (32 bit) from co-processor 0        */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MF     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00000  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint32_t xlr_read_32bit_cp0_desave(void)
{
  uint32_t val;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x4001f800  \n\t"  /* mfc0 at, $31 (desave), 0 (sel) */
      "nop               \n\t"
      "move   %0, $1     \n\t"

      ".set pop\n"

      : "=r" (val));

  return val;
}

/***************************************************************************/
/* xlr_read_64bit_cp0_desave() - Read desave (64 bit) from co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMF    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00001  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline uint64_t xlr_read_64bit_cp0_desave(void)
{
  uint32_t high, low;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      ".word 0x4021f800  \n\t"  /* dmfc0 at, $31 (desave), 0 (sel) */
      "nop               \n\t"
      "dsra32 %0, $1, 0  \n\t"
      "sll    %1, $1, 0  \n\t"

      ".set pop\n"

      : "=r" (high), "=r" (low)
  );

  return ( ((uint64_t)high) << 32) | low;
}




/***************************************************************************/
/* xlr_write_32bit_cp0_index() - Write index (32 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MT     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00100  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_32bit_cp0_index(uint32_t value)
{

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "move   $1, %0\n\t"

      ".word 0x40810000  \n\t"  /* mtc0 at, $0 (index), 0 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (value)
      : "$1"
  );
}

/***************************************************************************/
/* xlr_write_32bit_cp0_random() - Write random (32 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MT     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00100  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_32bit_cp0_random(uint32_t value)
{

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "move   $1, %0\n\t"

      ".word 0x40810800  \n\t"  /* mtc0 at, $1 (random), 0 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (value)
      : "$1"
  );
}

/***************************************************************************/
/* xlr_write_64bit_cp0_entrylo0() - Write entrylo0 (64 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMT    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00101  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_64bit_cp0_entrylo0(uint64_t value)
{
  uint32_t low, high;
  unsigned long flags;

  high = value >> 32;
  low  = value & 0xffffffff;

  __save_and_cli(flags);
  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "dsll32 $2, %1, 0  \n\t"
      "dsll32 $1, %0, 0  \n\t"
      "dsrl32 $2, $2, 0  \n\t"
      "or     $1, $1, $2 \n\t"

      ".word 0x40a11000  \n\t"  /* dmtc0 at, $2 (entrylo0), 0 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (high), "r" (low)
      : "$1", "$2"
  );
  __restore_flags(flags);
}

/***************************************************************************/
/* xlr_write_64bit_cp0_entrylo1() - Write entrylo1 (64 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMT    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00101  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_64bit_cp0_entrylo1(uint64_t value)
{
  uint32_t low, high;
  unsigned long flags;

  high = value >> 32;
  low  = value & 0xffffffff;

  __save_and_cli(flags);
  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "dsll32 $2, %1, 0  \n\t"
      "dsll32 $1, %0, 0  \n\t"
      "dsrl32 $2, $2, 0  \n\t"
      "or     $1, $1, $2 \n\t"

      ".word 0x40a11800  \n\t"  /* dmtc0 at, $3 (entrylo1), 0 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (high), "r" (low)
      : "$1", "$2"
  );
  __restore_flags(flags);
}

/***************************************************************************/
/* xlr_write_64bit_cp0_context() - Write context (64 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMT    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00101  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_64bit_cp0_context(uint64_t value)
{
  uint32_t low, high;

  high = value >> 32;
  low  = value & 0xffffffff;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "dsll32 $2, %1, 0  \n\t"
      "dsll32 $1, %0, 0  \n\t"
      "dsrl32 $2, $2, 0  \n\t"
      "or     $1, $1, $2 \n\t"

      ".word 0x40a12000  \n\t"  /* dmtc0 at, $4 (context), 0 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (high), "r" (low)
      : "$1", "$2"
  );
}

/***************************************************************************/
/* xlr_write_32bit_cp0_pagemask() - Write pagemask (32 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MT     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00100  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_32bit_cp0_pagemask(uint32_t value)
{

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "move   $1, %0\n\t"

      ".word 0x40812800  \n\t"  /* mtc0 at, $5 (pagemask), 0 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (value)
      : "$1"
  );
}

/***************************************************************************/
/* xlr_write_32bit_cp0_wired() - Write wired (32 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MT     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00100  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_32bit_cp0_wired(uint32_t value)
{

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "move   $1, %0\n\t"

      ".word 0x40810006  \n\t"  /* mtc0 at, $6 (wired), 0 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (value)
      : "$1"
  );
}

/***************************************************************************/
/* xlr_write_32bit_cp0_reserved0() - Write reserved0 (32 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MT     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00100  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_32bit_cp0_reserved0(uint32_t value)
{

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "move   $1, %0\n\t"

      ".word 0x40813800  \n\t"  /* mtc0 at, $7 (reserved), 0 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (value)
      : "$1"
  );
}

/***************************************************************************/
/* xlr_write_64bit_cp0_badvaddr() - Write badvaddr (64 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMT    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00101  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_64bit_cp0_badvaddr(uint64_t value)
{
  uint32_t low, high;

  high = value >> 32;
  low  = value & 0xffffffff;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "dsll32 $2, %1, 0  \n\t"
      "dsll32 $1, %0, 0  \n\t"
      "dsrl32 $2, $2, 0  \n\t"
      "or     $1, $1, $2 \n\t"

      ".word 0x40a14000  \n\t"  /* dmtc0 at, $8 (badvaddr), 0 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (high), "r" (low)
      : "$1", "$2"
  );
}

/***************************************************************************/
/* xlr_write_32bit_cp0_count() - Write count (32 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MT     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00100  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_32bit_cp0_count(uint32_t value)
{

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "move   $1, %0\n\t"

      ".word 0x40814800  \n\t"  /* mtc0 at, $9 (count), 0 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (value)
      : "$1"
  );
}

/***************************************************************************/
/* xlr_write_64bit_cp0_eirr() - Write eirr (64 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMT    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00101  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_64bit_cp0_eirr(uint64_t value)
{
  uint32_t low, high;

  high = value >> 32;
  low  = value & 0xffffffff;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "dsll32 $2, %1, 0  \n\t"
      "dsll32 $1, %0, 0  \n\t"
      "dsrl32 $2, $2, 0  \n\t"
      "or     $1, $1, $2 \n\t"

      ".word 0x40a14806  \n\t"  /* dmtc0 at, $9 (eirr), 6 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (high), "r" (low)
      : "$1", "$2"
  );
}

/***************************************************************************/
/* xlr_write_64bit_cp0_eimr() - Write eimr (64 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMT    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00101  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_64bit_cp0_eimr(uint64_t value)
{
  uint32_t low, high;

  high = value >> 32;
  low  = value & 0xffffffff;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "dsll32 $2, %1, 0  \n\t"
      "dsll32 $1, %0, 0  \n\t"
      "dsrl32 $2, $2, 0  \n\t"
      "or     $1, $1, $2 \n\t"

      ".word 0x40a14807  \n\t"  /* dmtc0 at, $9 (eimr), 7 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (high), "r" (low)
      : "$1", "$2"
  );
}


/***************************************************************************/
/* xlr_write_32bit_cp0_entryhi() - Write entryhi (32 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MT     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00100  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/

static inline void xlr_write_32bit_cp0_entryhi(uint32_t value)
{

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "move   $1, %0\n\t"

      ".word 0x40815000  \n\t"  /* mtc0 at, $10 (entryhi), 0 (sel) */
      "nop               \n\t"
      ".set pop\n"
      :
      : "r" (value)
      : "$1"
  );
}



/***************************************************************************/
/* xlr_write_64bit_cp0_entryhi() - Write entryhi (64 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMT    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00101  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_64bit_cp0_entryhi(uint64_t value)
{
  uint32_t low, high;

  high = value >> 32;
  low  = value & 0xffffffff;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "dsll32 $2, %1, 0  \n\t"
      "dsll32 $1, %0, 0  \n\t"
      "dsrl32 $2, $2, 0  \n\t"
      "or     $1, $1, $2 \n\t"

      ".word 0x40a15000  \n\t"  /* dmtc0 at, $10 (entryhi), 0 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (high), "r" (low)
      : "$1", "$2"
  );
}

/***************************************************************************/
/* xlr_write_32bit_cp0_compare() - Write compare (32 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MT     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00100  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_32bit_cp0_compare(uint32_t value)
{

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "move   $1, %0\n\t"

      ".word 0x40815800  \n\t"  /* mtc0 at, $11 (compare), 0 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (value)
      : "$1"
  );
}

/***************************************************************************/
/* xlr_write_32bit_cp0_status() - Write status (32 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MT     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00100  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_32bit_cp0_status(uint32_t value)
{

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "move   $1, %0\n\t"

      ".word 0x40816000  \n\t"  /* mtc0 at, $12 (status), 0 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (value)
      : "$1"
  );
}

/***************************************************************************/
/* xlr_write_32bit_cp0_cause() - Write cause (32 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MT     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00100  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_32bit_cp0_cause(uint32_t value)
{

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "move   $1, %0\n\t"

      ".word 0x40816800  \n\t"  /* mtc0 at, $13 (cause), 0 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (value)
      : "$1"
  );
}

/***************************************************************************/
/* xlr_write_32bit_cp0_epc() - Write epc (32 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MT     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00100  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_32bit_cp0_epc(uint32_t value)
{
   __asm__ __volatile__ (
	
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "move   $1, %0\n\t"

      ".word 0x40817000  \n\t"  /* mtc0 at, $14 (epc), 0 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (value)
      : "$1"
   );
}


/***************************************************************************/
/* xlr_write_64bit_cp0_epc() - Write epc (64 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMT    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00101  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_64bit_cp0_epc(uint64_t value)
{
  uint32_t low, high;

  high = value >> 32;
  low  = value & 0xffffffff;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "dsll32 $2, %1, 0  \n\t"
      "dsll32 $1, %0, 0  \n\t"
      "dsrl32 $2, $2, 0  \n\t"
      "or     $1, $1, $2 \n\t"

      ".word 0x40a17000  \n\t"  /* dmtc0 at, $14 (epc), 0 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (high), "r" (low)
      : "$1", "$2"
  );
}

/***************************************************************************/
/* xlr_write_32bit_cp0_processor_id() - Write processor_id (32 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MT     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00100  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_32bit_cp0_processor_id(uint32_t value)
{

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "move   $1, %0\n\t"

      ".word 0x40817800  \n\t"  /* mtc0 at, $15 (processor_id), 0 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (value)
      : "$1"
  );
}

/***************************************************************************/
/* xlr_write_32bit_cp0_config0() - Write config0 (32 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MT     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00100  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_32bit_cp0_config0(uint32_t value)
{

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "move   $1, %0\n\t"

      ".word 0x40818000  \n\t"  /* mtc0 at, $16 (config0), 0 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (value)
      : "$1"
  );
}

/***************************************************************************/
/* xlr_write_32bit_cp0_config1() - Write config1 (32 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MT     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00100  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_32bit_cp0_config1(uint32_t value)
{

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "move   $1, %0\n\t"

      ".word 0x40818001  \n\t"  /* mtc0 at, $16 (config1), 1 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (value)
      : "$1"
  );
}

/***************************************************************************/
/* xlr_write_64bit_cp0_watchlo() - Write watchlo (64 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMT    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00101  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_64bit_cp0_watchlo(uint64_t value)
{
  uint32_t low, high;

  high = value >> 32;
  low  = value & 0xffffffff;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "dsll32 $2, %1, 0  \n\t"
      "dsll32 $1, %0, 0  \n\t"
      "dsrl32 $2, $2, 0  \n\t"
      "or     $1, $1, $2 \n\t"

      ".word 0x40a19000  \n\t"  /* dmtc0 at, $18 (watchlo), 0 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (high), "r" (low)
      : "$1", "$2"
  );
}

/***************************************************************************/
/* xlr_write_32bit_cp0_watchhi() - Write watchhi (32 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MT     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00100  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_32bit_cp0_watchhi(uint32_t value)
{

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "move   $1, %0\n\t"

      ".word 0x40819800  \n\t"  /* mtc0 at, $19 (watchhi), 0 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (value)
      : "$1"
  );
}

/***************************************************************************/
/* xlr_write_64bit_cp0_xcontext() - Write xcontext (64 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMT    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00101  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_64bit_cp0_xcontext(uint64_t value)
{
  uint32_t low, high;

  high = value >> 32;
  low  = value & 0xffffffff;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "dsll32 $2, %1, 0  \n\t"
      "dsll32 $1, %0, 0  \n\t"
      "dsrl32 $2, $2, 0  \n\t"
      "or     $1, $1, $2 \n\t"

      ".word 0x40a1a000  \n\t"  /* dmtc0 at, $20 (xcontext), 0 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (high), "r" (low)
      : "$1", "$2"
  );
}

/***************************************************************************/
/* xlr_write_32bit_cp0_reserved1() - Write reserved1 (32 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MT     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00100  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_32bit_cp0_reserved1(uint32_t value)
{

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "move   $1, %0\n\t"

      ".word 0x4081a800  \n\t"  /* dmtc0 at, $21 (reserved), 0 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (value)
      : "$1"
  );
}

/***************************************************************************/
/* xlr_write_64bit_cp0_scratch0() - Write scratch0 (64 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMT    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00101  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_64bit_cp0_scratch0(uint64_t value)
{
  uint32_t low, high;

  high = value >> 32;
  low  = value & 0xffffffff;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "dsll32 $2, %1, 0  \n\t"
      "dsll32 $1, %0, 0  \n\t"
      "dsrl32 $2, $2, 0  \n\t"
      "or     $1, $1, $2 \n\t"

      ".word 0x40a1b000  \n\t"  /* dmtc0 at, $22 (scratch0), 0 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (high), "r" (low)
      : "$1", "$2"
  );
}

/***************************************************************************/
/* xlr_write_64bit_cp0_scratch1() - Write scratch1 (64 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMT    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00101  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_64bit_cp0_scratch1(uint64_t value)
{
  uint32_t low, high;

  high = value >> 32;
  low  = value & 0xffffffff;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "dsll32 $2, %1, 0  \n\t"
      "dsll32 $1, %0, 0  \n\t"
      "dsrl32 $2, $2, 0  \n\t"
      "or     $1, $1, $2 \n\t"

      ".word 0x40a1b001  \n\t"  /* dmtc0 at, $22 (scratch1), 1 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (high), "r" (low)
      : "$1", "$2"
  );
}

/***************************************************************************/
/* xlr_write_64bit_cp0_scratch2() - Write scratch2 (64 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMT    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00101  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_64bit_cp0_scratch2(uint64_t value)
{
  uint32_t low, high;

  high = value >> 32;
  low  = value & 0xffffffff;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "dsll32 $2, %1, 0  \n\t"
      "dsll32 $1, %0, 0  \n\t"
      "dsrl32 $2, $2, 0  \n\t"
      "or     $1, $1, $2 \n\t"

      ".word 0x40a1b002  \n\t"  /* dmtc0 at, $22 (scratch2), 2 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (high), "r" (low)
      : "$1", "$2"
  );
}

/***************************************************************************/
/* xlr_write_64bit_cp0_scratch3() - Write scratch3 (64 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMT    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00101  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_64bit_cp0_scratch3(uint64_t value)
{
  uint32_t low, high;

  high = value >> 32;
  low  = value & 0xffffffff;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "dsll32 $2, %1, 0  \n\t"
      "dsll32 $1, %0, 0  \n\t"
      "dsrl32 $2, $2, 0  \n\t"
      "or     $1, $1, $2 \n\t"

      ".word 0x40a1b003  \n\t"  /* dmtc0 at, $22 (scratch3), 3 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (high), "r" (low)
      : "$1", "$2"
  );
}

/***************************************************************************/
/* xlr_write_64bit_cp0_scratch4() - Write scratch4 (64 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMT    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00101  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_64bit_cp0_scratch4(uint64_t value)
{
  uint32_t low, high;

  high = value >> 32;
  low  = value & 0xffffffff;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "dsll32 $2, %1, 0  \n\t"
      "dsll32 $1, %0, 0  \n\t"
      "dsrl32 $2, $2, 0  \n\t"
      "or     $1, $1, $2 \n\t"

      ".word 0x40a1b004  \n\t"  /* dmtc0 at, $22 (scratch4), 4 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (high), "r" (low)
      : "$1", "$2"
  );
}

/***************************************************************************/
/* xlr_write_64bit_cp0_scratch5() - Write scratch5 (64 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMT    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00101  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_64bit_cp0_scratch5(uint64_t value)
{
  uint32_t low, high;

  high = value >> 32;
  low  = value & 0xffffffff;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "dsll32 $2, %1, 0  \n\t"
      "dsll32 $1, %0, 0  \n\t"
      "dsrl32 $2, $2, 0  \n\t"
      "or     $1, $1, $2 \n\t"

      ".word 0x40a1b005  \n\t"  /* dmtc0 at, $22 (scratch5), 5 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (high), "r" (low)
      : "$1", "$2"
  );
}

/***************************************************************************/
/* xlr_write_64bit_cp0_scratch6() - Write scratch6 (64 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMT    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00101  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_64bit_cp0_scratch6(uint64_t value)
{
  uint32_t low, high;

  high = value >> 32;
  low  = value & 0xffffffff;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "dsll32 $2, %1, 0  \n\t"
      "dsll32 $1, %0, 0  \n\t"
      "dsrl32 $2, $2, 0  \n\t"
      "or     $1, $1, $2 \n\t"

      ".word 0x40a1b006  \n\t"  /* dmtc0 at, $22 (scratch6), 6 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (high), "r" (low)
      : "$1", "$2"
  );
}

/***************************************************************************/
/* xlr_write_64bit_cp0_scratch7() - Write scratch7 (64 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMT    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00101  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_64bit_cp0_scratch7(uint64_t value)
{
  uint32_t low, high;

  high = value >> 32;
  low  = value & 0xffffffff;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "dsll32 $2, %1, 0  \n\t"
      "dsll32 $1, %0, 0  \n\t"
      "dsrl32 $2, $2, 0  \n\t"
      "or     $1, $1, $2 \n\t"

      ".word 0x40a1b007  \n\t"  /* dmtc0 at, $22 (scratch7), 7 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (high), "r" (low)
      : "$1", "$2"
  );
}

/***************************************************************************/
/* xlr_write_32bit_cp0_debug() - Write debug (32 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MT     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00100  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_32bit_cp0_debug(uint32_t value)
{

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "move   $1, %0\n\t"

      ".word 0x4081b800  \n\t"  /* mtc0 at, $23 (debug), 0 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (value)
      : "$1"
  );
}

/***************************************************************************/
/* xlr_write_32bit_cp0_depc() - Write depc (32 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MT     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00100  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_32bit_cp0_depc(uint32_t value)
{

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "move   $1, %0\n\t"

      ".word 0x4081c000  \n\t"  /* mtc0 at, $24 (depc), 0 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (value)
      : "$1"
  );
}

/***************************************************************************/
/* xlr_write_32bit_cp0_perfctr0_lo() - Write perfctr0_lo (32 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MT     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00100  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_32bit_cp0_perfctr0_lo(uint32_t value)
{

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "move   $1, %0\n\t"

      ".word 0x4081c800  \n\t"  /* mtc0 at, $25 (perfctr0), 0 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (value)
      : "$1"
  );
}

/***************************************************************************/
/* xlr_write_32bit_cp0_perfctr0_high() - Write perfctr0_high (32 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MT     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00100  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_32bit_cp0_perfctr0_high(uint32_t value)
{

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "move   $1, %0\n\t"

      ".word 0x4081c801  \n\t"  /* mtc0 at, $25 (perfctr0), 1 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (value)
      : "$1"
  );
}

/***************************************************************************/
/* xlr_write_32bit_cp0_perfctr1_lo() - Write perfctr1_lo (32 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MT     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00100  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_32bit_cp0_perfctr1_lo(uint32_t value)
{

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "move   $1, %0\n\t"

      ".word 0x4081c802  \n\t"  /* mtc0 at, $25 (perfctr1), 2 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (value)
      : "$1"
  );
}

/***************************************************************************/
/* xlr_write_32bit_cp0_perfctr1_high() - Write perfctr1_high (32 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MT     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00100  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_32bit_cp0_perfctr1_high(uint32_t value)
{

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "move   $1, %0\n\t"

      ".word 0x4081c803  \n\t"  /* mtc0 at, $25 (perfctr1), 3 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (value)
      : "$1"
  );
}

/***************************************************************************/
/* xlr_write_32bit_cp0_errorctl() - Write errorctl (32 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   MT     |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00100  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_32bit_cp0_errorctl(uint32_t value)
{

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "move   $1, %0\n\t"

      ".word 0x4081d000  \n\t"  /* mtc0 at, $26 (errorctl), 0 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (value)
      : "$1"
  );
}



/***************************************************************************/
/* xlr_write_64bit_cp0_taglo0() - Write taglo0 (64 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMT    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00101  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_64bit_cp0_taglo0(uint64_t value)
{
  uint32_t low, high;

  high = value >> 32;
  low  = value & 0xffffffff;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "dsll32 $2, %1, 0  \n\t"
      "dsll32 $1, %0, 0  \n\t"
      "dsrl32 $2, $2, 0  \n\t"
      "or     $1, $1, $2 \n\t"

      ".word 0x40a1e000  \n\t"  /* dmtc0 at, $28 (taglo), 0 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (high), "r" (low)
      : "$1", "$2"
  );
}


/***************************************************************************/
/* xlr_write_64bit_cp0_taglo1() - Write taglo1 (64 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMT    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00101  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_64bit_cp0_taglo1(uint64_t value)
{
  uint32_t low, high;

  high = value >> 32;
  low  = value & 0xffffffff;

  __asm__ __volatile__ (
    ".set push\n"
    ".set noreorder\n"
    ".set noat\n"
    ".set mips4\n"

    "dsll32 $2, %1, 0  \n\t"
    "dsll32 $1, %0, 0  \n\t"
    "dsrl32 $2, $2, 0  \n\t"
    "or     $1, $1, $2 \n\t"
     
    ".word 0x40a1e001  \n\t"  /* dmtc0 at, $28 (taglo), 1 (sel) */
    "nop               \n\t"
     
    ".set pop\n"
      
    :
    : "r" (high), "r" (low)
    : "$1", "$2"
 );
}


/***************************************************************************/
/* xlr_write_64bit_cp0_taglo2() - Write taglo2 (64 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMT    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00101  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_64bit_cp0_taglo2(uint64_t value)
{
  uint32_t low, high;

  high = value >> 32;
  low  = value & 0xffffffff;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "dsll32 $2, %1, 0  \n\t"
      "dsll32 $1, %0, 0  \n\t"
      "dsrl32 $2, $2, 0  \n\t"
      "or     $1, $1, $2 \n\t"

      ".word 0x40a1e002  \n\t"  /* dmtc0 at, $28 (taglo), 2 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (high), "r" (low)
      : "$1", "$2"
  );
}



/***************************************************************************/
/* xlr_write_64bit_cp0_taghi0() - Write taghi0 (64 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMT    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00101  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_64bit_cp0_taghi0(uint64_t value)
{
  uint32_t low, high;

  high = value >> 32;
  low  = value & 0xffffffff;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "dsll32 $2, %1, 0  \n\t"
      "dsll32 $1, %0, 0  \n\t"
      "dsrl32 $2, $2, 0  \n\t"
      "or     $1, $1, $2 \n\t"

      ".word 0x40a1e800  \n\t"  /* dmtc0 at, $29 (taghi), 0 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (high), "r" (low)
      : "$1", "$2"
  );
}


/***************************************************************************/
/* xlr_write_64bit_cp0_taghi1() - Write taghi1 (64 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMT    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00101  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_64bit_cp0_taghi1(uint64_t value)
{
	uint32_t low, high;

	high = value >> 32;
	low  = value & 0xffffffff;
	
      __asm__ __volatile__ (
	".set push\n"
	".set noreorder\n"
      	".set noat\n"
      	".set mips4\n"
      
	"dsll32 $2, %1, 0  \n\t"
	"dsll32 $1, %0, 0  \n\t"
	"dsrl32 $2, $2, 0  \n\t"
	"or     $1, $1, $2 \n\t"
			  
	".word 0x40a1e801  \n\t"  /* dmtc0 at, $29 (taghi), 1 (sel) */
	"nop               \n\t"
	      
	".set pop\n"

	:
	: "r" (high), "r" (low)
	: "$1", "$2"
  );
}
		    

/***************************************************************************/
/* xlr_write_64bit_cp0_taghi2() - Write taghi2 (64 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMT    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00101  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_64bit_cp0_taghi2(uint64_t value)
{
  uint32_t low, high;

  high = value >> 32;
  low  = value & 0xffffffff;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "dsll32 $2, %1, 0  \n\t"
      "dsll32 $1, %0, 0  \n\t"
      "dsrl32 $2, $2, 0  \n\t"
      "or     $1, $1, $2 \n\t"

      ".word 0x40a1e802  \n\t"  /* dmtc0 at, $29 (taghi), 2 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (high), "r" (low)
      : "$1", "$2"
  );
}


/***************************************************************************/
/* xlr_write_64bit_cp0_errorepc() - Write errorepc (64 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMT    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00101  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_64bit_cp0_errorepc(uint64_t value)
{
  uint32_t low, high;

  high = value >> 32;
  low  = value & 0xffffffff;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "dsll32 $2, %1, 0  \n\t"
      "dsll32 $1, %0, 0  \n\t"
      "dsrl32 $2, $2, 0  \n\t"
      "or     $1, $1, $2 \n\t"

      ".word 0x40a1f000  \n\t"  /* dmtc0 at, $30 (errorepc), 0 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (high), "r" (low)
      : "$1", "$2"
  );
}

/***************************************************************************/
/* xlr_write_64bit_cp0_desave() - Write desave (64 bit) to co-processor 0              */
/*                                                                         */
/* Opcode for:                                                             */
/*                                                                         */
/* mfc0  rt, rd, sel                                                       */
/*                                                                         */
/* 31    26  25     21  20    16  15    11 10             3  2   0         */
/* ---------------------------------------------------------------         */
/* |        |          |         |        |                |     |         */
/* | COP0   |   DMT    |   rt    |   rd   |       0        | sel |         */
/* | 010000 |   00101  |   rt    |   rd   |    00000000    | sel |         */
/* ---------------------------------------------------------------         */
/*     6          5        5         5            8           3            */
/***************************************************************************/
static inline void xlr_write_64bit_cp0_desave(uint64_t value)
{
  uint32_t low, high;

  high = value >> 32;
  low  = value & 0xffffffff;

  __asm__ __volatile__ (
      ".set push\n"
      ".set noreorder\n"
      ".set noat\n"
      ".set mips4\n"

      "dsll32 $2, %1, 0  \n\t"
      "dsll32 $1, %0, 0  \n\t"
      "dsrl32 $2, $2, 0  \n\t"
      "or     $1, $1, $2 \n\t"

      ".word 0x40a1f800  \n\t"  /* dmtc0 at, $31 (desave), 0 (sel) */
      "nop               \n\t"

      ".set pop\n"

      :
      : "r" (high), "r" (low)
      : "$1", "$2"
  );
}

static __inline__ unsigned int read_32bit_phnx_ctrl_reg(int block, int reg)
{
	  unsigned int __res;

	  __asm__ __volatile__(

			   ".set\tpush\n\t"
			   ".set\tnoreorder\n"
			   "move $9, %1\n"
			   ".word 0x71280018\n"
                           "move %0, $8\n"
                           ".set\tpop"
			   : "=r" (__res) : "r"((block<<8)|reg)
			   : "$8", "$9"
 	);

	return __res;
}

static __inline__ void write_32bit_phnx_ctrl_reg(int block, int reg, unsigned int value)
{
	  __asm__ __volatile__(
		".set\tpush\n\t"
		".set\tnoreorder\n\t"
		"move $8, %0\n"
		"move $9, %1\n"
		".word 0x71280019\n"
		".set\tpop"
		:
		:"r" (value), "r"((block<<8)|reg)
		: "$8", "$9"
	);
}

#define read_phnx_ctrl(addr)                                  \
({unsigned long long __res;                                   \
	__asm__ __volatile__ (                                \
	      "mfcr %0, %1\n" : "=r"(__res) : "r"(addr));     \
	       __res;})
 
#define write_phnx_ctrl(addr, value)                          \
         __asm__ __volatile__ (                               \
			         "mtcr %0, %1\n": : "r" (value), "r" (addr));

#if (_MIPS_SIM == _MIPS_SIM_ABI64)
  #define PTR2U64(x) (unsigned long)(x)
#else /* preserve sign-extension */
  #define PTR2U64(x) (int)(x)
#endif

/*
 * the following macros are imported from the previous incarnation
 * of asmmacros.h. Some of them are subsumed by the macros defined
 * above and need to be elided at some point. A few macros are not
 * defined above and hence need to be retained.
 */

/*
 * Macros to access the system control coprocessor
 */

#define read_32bit_cp0_register(source) \
({ int __res; \
    __asm__ __volatile__( \
	".set\tpush\n\t" \
	".set\tnoreorder\n\t" \
    "mfc0\t%0,"STR(source)"\n\t" \
	".set\tpop"	\
    : "=r" (__res)); \
    __res;})

#define write_32bit_cp0_register(register,value) \
    __asm__ __volatile__( \
    "mtc0\t%0,"STR(register)"\n\t" \
	"nop" \
    : : "r" (value));

#define read_32bit_cp1_ctrl_register(source) \
({ int __res; \
    __asm__ __volatile__( \
        ".set \tpush\n\t" \
        ".set \treorder\n\t" \
        "cfc1 \t%0,"STR(source)"\n\t" \
        ".set \tpop" \
        : "=r" (__res)); \
        __res;})

#define write_32bit_cp1_ctrl_register(register,value) \
    __asm__ __volatile__( \
        ".set \tpush\n\t" \
	".set noat\n\t"	     \
        "move $1, %0 \n\t" \
        "ctc1\t$1,"STR(register)"\n\t" \
        ".set \tpop\n\t" \
        "nop" \
        : : "r" (value));

#define read_32bit_cp2_register(source) \
({ int __res; \
    __asm__ __volatile__( \
        ".set \tpush\n\t" \
        ".set \treorder\n\t" \
        "mfc2 \t%0,"STR(source)"\n\t" \
        ".set \tpop" \
        : "=r" (__res)); \
        __res;})

#define write_32bit_cp2_register(register,value) \
    __asm__ __volatile__( \
        "mtc2\t%0,"STR(register)"\n\t" \
        "nop" \
        : : "r" (value));

#define read_32bit_cp2_register_sel(source, sel) \
({ int __res; \
    __asm__ __volatile__( \
        ".set\tpush\n\t" \
        ".set mips32\n\t" \
        "mfc2\t%0,"STR(source)", %1\n\t" \
        ".set\tpop" \
        : "=r" (__res) : "i" (sel) ); \
        __res;})

#define write_32bit_cp2_register_sel(reg, value, sel) \
    __asm__ __volatile__( \
        ".set\tpush\n\t" \
        ".set mips32\n\t" \
        "mtc2\t%0,"STR(reg)", %1\n\t" \
        ".set\tpop" \
        : : "r" (value), "i" (sel) );        

#define read_64bit_cp2_register(source) \
({ unsigned long long __res; \
            __asm__ __volatile__( \
    ".set\tpush\n\t" \
    ".set\treorder\n\t" \
            ".set\tmips64\n\t" \
            "dmfc2\t%0,"STR(source)"\n\t" \
    ".set\tpop" \
            : "=r" (__res)); \
            __res;})

#define write_64bit_cp2_register(register,value) \
            __asm__ __volatile__( \
    ".set\tpush\n\t" \
    ".set\treorder\n\t" \
            "dmtc2\t%0,"STR(register)"\n\t" \
    "nop" \
    ".set\tpop" \
            : : "r" (value));

#define read_64bit_cp2_register_sel(source, sel) \
({ unsigned long long __res; \
            __asm__ __volatile__( \
    ".set\tpush\n\t" \
            ".set mips64\n\t" \
            "dmfc2\t%0,"STR(source)", %1\n\t" \
    ".set\tpop" \
            : "=r" (__res) : "i" (sel) ); \
            __res;})

#define write_64bit_cp2_register_sel(reg, value, sel) \
            __asm__ __volatile__( \
    ".set\tpush\n\t" \
            ".set mips64\n\t" \
            "dmtc2\t%0,"STR(reg)", %1\n\t" \
    ".set\tpop" \
            : : "r" (value), "i" (sel) );    

/*
 * This should be changed when we get a compiler that support the MIPS32 ISA.
 */

static inline void set_entrylo0(unsigned long long val)
{
	unsigned long flags;

	__save_and_cli(flags);
	__asm__ __volatile__(
		".set\tmips3\n\t"
		"dsll\t%L0, %L0, 32\n\t"
		"dsrl\t%L0, %L0, 32\n\t"
		"dsll\t%M0, %M0, 32\n\t"
		"or\t%L0, %L0, %M0\n\t"
		"dmtc0\t%L0, $2\n\t"
		".set\tmips0"
		: : "r" (val));
	__restore_flags(flags);
}

static inline void set_entrylo1(unsigned long long val)
{
	unsigned long flags;

	__save_and_cli(flags);
	__asm__ __volatile__(
		".set\tmips3\n\t"
		"dsll\t%L0, %L0, 32\n\t"
		"dsrl\t%L0, %L0, 32\n\t"
		"dsll\t%M0, %M0, 32\n\t"
		"or\t%L0, %L0, %M0\n\t"
		"dmtc0\t%L0, $3\n\t"
		".set\tmips0"
		: : "r" (val));
	__restore_flags(flags);
}



static inline void set_entryhi(unsigned long val)
{
	__asm__ __volatile__(
		".set push\n\t"
		".set reorder\n\t"
		"mtc0 %z0, $10\n\t"
		".set pop"
		: : "Jr" (val));
}

static inline unsigned long get_entryhi(void)
{
	unsigned long val;

	__asm__ __volatile__(
		".set push\n\t"
		".set reorder\n\t"
		"mfc0 %0, $10\n\t"
		".set pop"
		: "=r" (val));

	return val;
}

static inline void set_index(unsigned long val)
{
	__asm__ __volatile__(
		".set push\n\t"
		".set reorder\n\t"
		"mtc0 %z0, $0\n\t"
		".set pop"
		: : "Jr" (val));
}

static inline unsigned long get_index(void)
{
	unsigned long val;

	__asm__ __volatile__(
		".set push\n\t"
		".set reorder\n\t"
		"mfc0 %0, $0\n\t"
		".set pop"
		: "=r" (val));
	return val;
}

/* TLB operations. */
static inline void tlb_probe(void)
{
	__asm__ __volatile__(
		".set push\n\t"
		".set reorder\n\t"
		"tlbp\n\t"
		".set pop");
}

static inline void tlb_read(void)
{
	__asm__ __volatile__(
		".set push\n\t"
		".set reorder\n\t"
		"tlbr\n\t"
		".set pop");
}

static inline void tlb_write_indexed(void)
{
	__asm__ __volatile__(
		".set push\n\t"
		".set reorder\n\t"
		"tlbwi\n\t"
		".set pop");
}

static inline void tlb_write_random(void)
{
	__asm__ __volatile__(
		".set push\n\t"
		".set reorder\n\t"
		"tlbwr\n\t"
		".set pop");
}

#define  DMTC0_AT_scratch4        0x40a1b004
#define  DMTC0_AT_scratch5        0x40a1b005
#define  MFC0_AT_index            0x40010000
#define  DMTC0_AT_scratch6        0x40a1b006
#define  DMTC0_AT_scratch7        0x40a1b007
#define  MTC0_AT_depc             0x4081c000
#define  DMFC0_AT_scratch0        0x4021b000
#define  DMFC0_AT_epc             0x40217000
#define  DMFC0_AT_scratch1        0x4021b001
#define  DMFC0_AT_scratch2        0x4021b002
#define  DMFC0_AT_scratch3        0x4021b003
#define  DMFC0_AT_scratch4        0x4021b004
#define  DMFC0_AT_scratch5        0x4021b005
#define  DMFC0_AT_scratch6        0x4021b006
#define  DMFC0_AT_scratch7        0x4021b007
#define  MTC0_AT_perfctr1_lo      0x4081c802
#define  DMFC0_AT_eirr            0x40214806
#define  MFC0_AT_reserved0        0x40013800
#define  MFC0_AT_reserved2        0x4001a800
#define  MTC0_AT_cacheerr         0x4081d800
#define  MTC0_AT_count            0x40814800
#define  DMTC0_AT_entryhi         0x40a15000
#define  DMTC0_AT_context         0x40a12000
#define  MFC0_AT_perfctr0_hi      0x4001c801
#define  MFC0_AT_cacheerr         0x4001d800
#define  DMTC0_AT_eimr            0x40a14807
#define  MFC0_AT_wired            0x40010006
#define  DMFC0_AT_context         0x40212000
#define  MTC0_AT_perfctr0_lo      0x4081c800
#define  MTC0_AT_reserved0        0x40813800
#define  MTC0_AT_reserved1        0x4081a800
#define  MFC0_AT_desave           0x4001f800
#define  DMTC0_AT_badvaddr        0x40a14000
#define  MTC0_AT_cause            0x40816800
#define  MTC0_AT_random           0x40810800
#define  MFC0_AT_random           0x40010800
#define  MFC0_AT_debug            0x4001b800
#define  MFC0_AT_depc             0x4001c000
#define  DMFC0_AT_lladdr          0x40218800
#define  MTC0_AT_config0          0x40818000
#define  MTC0_AT_config1          0x40818001
#define  DMFC0_AT_entryhi         0x40215000
#define  MTC0_AT_perfctr1_high    0x4081c803
#define  DMFC0_AT_eimr            0x40214807
#define  MTC0_AT_index            0x40810000
#define  DMTC0_AT_epc             0x40a17000
#define  DMFC0_AT_tahlo2          0x4021e002
#define  MFC0_AT_diagnostic       0x4001b000
#define  DMTC0_AT_taghi0          0x40a1e800
#define  DMFC0_AT_taghi0          0x4021e800
#define  DMFC0_AT_badvaddr        0x40214000
#define  MFC0_AT_count            0x40014800
#define  DMTC0_AT_taghi2          0x40a1e802
#define  DMTC0_AT_entrylo0        0x40a11000
#define  DMTC0_AT_entrylo1        0x40a11800
#define  DMFC0_AT_taghi2          0x4021e802
#define  MTC0_AT_compare          0x40815800
#define  MFC0_AT_perfctr1_high    0x4001c803
#define  DMTC0_AT_taglo0          0x40a1e000
#define  DMFC0_AT_taglo0          0x4021e000
#define  DMFC0_AT_reserved1       0x4021a800
#define  DMTC0_AT_taglo2          0x40a1e002
#define  MTC0_AT_errorctl         0x4081d000
#define  MTC0_AT_pagemask         0x40812800
#define  MTC0_AT_watchhi          0x40819800
#define  DMTC0_AT_watchlo         0x40a19000
#define  DMFC0_AT_entrylo0        0x40211000
#define  DMFC0_AT_entrylo1        0x40211800
#define  DMTC0_AT_xcontext        0x40a1a000
#define  MFC0_AT_perfctr0_low     0x4001c800
#define  MTC0_AT_perfctr0_high    0x4081c801
#define  MFC0_AT_config0          0x40018000
#define  MFC0_AT_conf             0x40011800
#define  MFC0_AT_errorctl         0x4001d000
#define  MFC0_AT_config1          0x40018001
#define  MFC0_AT_processor_id     0x40017800
#define  MFC0_AT_compare          0x40015800
#define  DMTC0_AT_desave          0x40a1f800
#define  DMFC0_AT_desave          0x4021f800
#define  DMTC0_AT_errorepc        0x40a1f000
#define  DMFC0_AT_watchlo         0x40219000
#define  MFC0_AT_watchhi          0x40019800
#define  MTC0_AT_wired            0x40810006
#define  DMFC0_AT_errorepc        0x4021f000
#define  MFC0_AT_pagemask         0x40012800
#define  MFC0_AT_perfctr1_low     0x4001c802
#define  DMFC0_AT_xcontext        0x4021a000
#define  MTC0_AT_processor_id     0x40817800
#define  MTC0_AT_status           0x40816000
#define  MFC0_AT_status           0x40016000
#define  MTC0_AT_debug            0x4081b800
#define  DMTC0_AT_eirr            0x40a14806
#define  MFC0_AT_cause            0x40016800
#define  DMTC0_AT_scratch0        0x40a1b000
#define  DMTC0_AT_scratch1        0x40a1b001
#define  DMTC0_AT_scratch2        0x40a1b002
#define  DMTC0_AT_scratch3        0x40a1b003

#else /* ifndef __ASSEMBLY__ */

/* Read the cpu id from the cp0 config register
 * cpuid[9:4], thrid[3: 0]
 */

.macro phoenix_cpu_id id, temp0
    /* get the cpu id */
    mfc0    \id, CP0_CONFIG, 7
	srl     \id, \id, 4
	andi    \id, \id, 0x3f
	sll     \id, \id, 2
		
	/* get the thread id */
	mfc0    \temp0, CP0_CONFIG, 7
    andi    \temp0, 0x0f

	/* calculate the linear cpu id */
    addu    \id, \id, \temp0
.endm

#endif

#endif
