/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/

#ifndef _ELF_EXT_H
#define _ELF_EXT_H

#include "xlr_boot_lib.h"
#include "elf.h"

#define ELF32_GET_ADDR(ptr) (*(int *)ptr)
#define ELF64_GET_ADDR(ptr) (*ptr)

extern void  elf_initialize(void);
extern int   elfload_shell_cmd(int argc, char *argv[]);

#define MAX_ELF_SECTIONS  16
#define MAX_ELF_SEGMENTS  16

#define init_done(segment)      (segment->init_done)
#define is_shared(segment)      (segment->is_shared)

#define segment_size(elf, seg) ((memsz (elf, seg) + align (elf, seg) - 1) & ~(align (elf, seg) - 1))

#define vaddr(elf, segment)             (elf_uval (elf, &(segment->hdr), p_vaddr))
#define paddr(elf, segment)             (elf_uval (elf, &(segment->hdr), p_paddr))
#define align(elf, segment)             (elf_uval (elf, &(segment->hdr), p_align))
#define offset(elf, segment)            (elf_uval (elf, &(segment->hdr), p_offset))
#define filesz(elf, segment)            (elf_uval (elf, &(segment->hdr), p_filesz))
#define memsz(elf, segment)             (elf_uval (elf, &(segment->hdr), p_memsz))

#define is_code_segment(elf, phdr)      (elf_uval (elf, phdr, p_flags) & PF_X)

typedef struct {
        elf_phdr hdr;   //Copy of phdr element
        int    init_done;
        int    is_shared;
        int    ntlb;
        struct cpu_tlb_mapping map[MAX_TLB_MAPPINGS];
} Elf_ExtPhdr;

#endif
