/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/

#ifndef _FIFO_H
#define _FIFO_H

#define FIFO_SIZE 2048

struct fifo {
  int data[FIFO_SIZE];
  int size;
  int head;
  int tail;
};

static __inline__ int  fifo_next_head(struct fifo *fifo) 
{ 
  return (fifo->head+1) % fifo->size ; 
}

static __inline__ int  fifo_next_tail(struct fifo *fifo) 
{ 
  return (fifo->tail+1) % fifo->size ; 
}

static __inline__ int  fifo_empty(struct fifo *fifo) 
{ 
  return fifo->head == fifo->tail ? 1 : 0; 
}

static __inline__ int  fifo_full(struct fifo *fifo) 
{ 
  return fifo_next_tail(fifo) == fifo->head ? 1 : 0 ; 
}

static __inline__ int  fifo_size(struct fifo *fifo) 
{ 
  if (fifo->head <= fifo->tail)
    return fifo->tail - fifo->head;
  else
    return (fifo->size - fifo->head) + (fifo->tail - 0);
}

static __inline__ int  fifo_dequeue(struct fifo *fifo, int *data) 
{
  if (fifo_empty(fifo))
    return 0;
  *data = (fifo->data)[fifo->head];
  fifo->head = fifo_next_head(fifo);
  return 1;
}

static __inline__ int fifo_enqueue(struct fifo *fifo, int data) 
{
  if (fifo_full(fifo))
    return 0;
  fifo->data[fifo->tail] = data;
  fifo->tail = fifo_next_tail(fifo);
  return 1;
}

static __inline__ void fifo_init(struct fifo *f)
{
  f->head = f->tail = 0;
  f->size = FIFO_SIZE;
}

#endif
