/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/

#ifndef _MIPS_EXTS_H
#define _MIPS_EXTS_H

#include "mipsregs.h"

#ifndef __ASSEMBLY__

#define read_64bit_cp0_register_sel(source, sel)                \
({ unsigned long long __res;                                    \
        __asm__ __volatile__(                                   \
			".set\tpush\n\t"									\
			".set mips64\n\t"									\
			"dmfc0\t%0,"STR(source)", %1\n\t"					\
			".set\tpop"											\
			: "=r" (__res) : "i" (sel) );						\
        __res;})

#define write_64bit_cp0_register_sel(reg, value, sel)           \
        __asm__ __volatile__(                                   \
			".set\tpush\n\t"									\
			".set mips64\n\t"									\
			"dmtc0\t%0,"STR(reg)", %1\n\t"						\
			"nop\n\t"											\
			"nop\n\t"											\
			"nop\n\t"											\
			"nop\n\t"											\
			".set\tpop"											\
			: : "r" (value), "i" (sel) );                          

#define read_32bit_cp0_register_sel(source, sel)                \
	({ unsigned int __res;										\
        __asm__ __volatile__(                                   \
			".set\tpush\n\t"									\
			".set mips32\n\t"									\
			"mfc0\t%0,"STR(source)", %1\n\t"					\
			".set\tpop"											\
			: "=r" (__res) : "i" (sel) );						\
        __res;})

#define write_32bit_cp0_register_sel(reg, value, sel)	\
	__asm__ __volatile__(								\
		".set\tpush\n\t"								\
        ".set mips32\n\t"								\
        "mtc0\t%0,"STR(reg)", %1\n\t"					\
		".set\tpop"										\
        : : "r" (value), "i" (sel) );                          

#define read_c0_count()     read_32bit_cp0_register_sel(CP0_COUNT, 0)

#define read_c0_compare()     read_32bit_cp0_register_sel(CP0_COMPARE, 0)
#define write_c0_compare(val) write_32bit_cp0_register_sel(CP0_COMPARE, val, 0)

#define read_phnx_eirr() read_64bit_cp0_register_sel(CP0_COUNT, 6)
#define write_phnx_eirr(val) write_64bit_cp0_register_sel(CP0_COUNT, val, 6)

#define read_phnx_eimr() read_64bit_cp0_register_sel(CP0_COUNT, 7)
#define write_phnx_eimr(val) write_64bit_cp0_register_sel(CP0_COUNT, val, 7)

#define read_c0_status()  read_32bit_cp0_register_sel(CP0_STATUS, 0)
#define write_c0_status(val) write_32bit_cp0_register_sel(CP0_STATUS, val, 0)

#define read_c0_ebase() read_32bit_cp0_register_sel(CP0_EBASE, 1)
#define write_c0_ebase(val) write_32bit_cp0_register_sel(CP0_EBASE, val, 1)

#define read_c0_index() read_32bit_cp0_register_sel(CP0_INDEX, 0)
#define write_c0_index(val) write_32bit_cp0_register_sel(CP0_INDEX, val, 0)

#define read_c0_wired() read_32bit_cp0_register_sel(CP0_WIRED, 0)
#define write_c0_wired(val) write_32bit_cp0_register_sel(CP0_WIRED, val, 0)

#define write_c0_pagemask(val) write_32bit_cp0_register_sel(CP0_PAGEMASK, val, 0)
#define read_c0_config1() read_32bit_cp0_register_sel(CP0_CONFIG,1)
#define write_c0_entryhi(val) write_64bit_cp0_register_sel(CP0_ENTRYHI, val, 0)
#define write_c0_entrylo0(val) write_64bit_cp0_register_sel(CP0_ENTRYLO0, val, 0)
#define write_c0_entrylo1(val) write_64bit_cp0_register_sel(CP0_ENTRYLO1, val, 0)

/*******************************************************************/
#ifdef XLR
static __inline__ int ldadd_w(unsigned int value, int *addr)
{
  __asm__ __volatile__(
		       ".set push\n"
		       ".set noreorder\n"
		       "move $8, %2\n"
		       "move $9, %3\n"
		       //"ldaddw $8, $9\n"
		       ".word 0x71280010\n"
		       "move %0, $8\n"
		       ".set pop\n"
		       :"=&r"(value), "+m"(*addr)
		       : "0" (value), "r"((unsigned long)addr)
		       : "$8", "$9"
		       );
  return value;
}

static __inline__ unsigned int ldadd_wu(unsigned int value, unsigned int *addr)
{
  __asm__ __volatile__(
		       ".set push\n"
		       ".set noreorder\n"
		       "move $8, %2\n"
		       "move $9, %3\n"
		       //"ldaddwu $8, $9\n"
		       ".word 0x71280011\n"
		       "move %0, $8\n"
		       ".set pop\n"
		       :"=&r"(value), "+m"(*addr)
		       : "0"(value), "r"((unsigned long)addr)
		       : "$8", "$9"
		       );
  return value;
}

// reads the old value
static __inline__ long long ldadd_d(long long value, long long *addr)
{
  __asm__ __volatile__(
		       ".set push\n"
		       ".set noreorder\n"
		       "move $8, %2\n"
		       "move $9, %3\n"
		       "ldaddd $8, $9\n"
		       "move %0, $8\n"
		       ".set pop\n"
		       :"=&r"(value), "+m"(*addr)
		       : "0" (value), "r"((unsigned long)addr)
		       : "$8", "$9"
		       );
  return value;
}

// doesn't read the old value
static __inline__ void ldadd_d_noread(long long value, long long *addr)
{
  __asm__ __volatile__(
		       ".set push\n"
		       ".set noreorder\n"
		       "move $8, %2\n"
		       "move $9, %3\n"
		       "ldaddd $8, $9\n"
		       // "move %0, $8\n"
		       ".set pop\n"
		       : "=&r"(value), "+m"(*addr)
		       : "0" (value), "r"((unsigned long)addr)
		       : "$8", "$9"
		       );
}
#endif /* XLR */

static __inline__ int processor_id(void)
{
	int cpu;

	__asm__ __volatile__ (
		".set push\n"
		".set noreorder\n"
		".set mips32\n"
		"mfc0 $8, $15, 1\n"
		"andi %0, $8, 0x3ff\n"
		".set pop\n"
		: "=&r"(cpu) : : "$8"
		);

	return cpu;
}

static __inline__ int phnx_cpu_id(void)
{
	  unsigned int cpu=0;

	    __asm__ volatile (
		".set push\n"
		".set mips32\n"
		"mfc0 %0, $15, 1\n"
		".set pop\n"
		:"=r" (cpu)
	);
	return ((cpu >> 2) & 0x7);
}

static __inline__ int phnx_thr_id(void)
{
	  unsigned int cpu=0;

	    __asm__ volatile (
		".set push\n"
		".set mips32\n"
		"mfc0 %0, $15, 1\n"
		".set pop\n"
		:"=r" (cpu)
	   );
	   return (cpu & 0x03);
}

extern unsigned long tlb_size_to_page_size(unsigned long size);
extern unsigned long tlb_size_to_mask(unsigned long size);

#endif

#endif
