/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/

// top level configuration structure
// for all networking devices
// gmac, xgmac, spi4
//

#ifndef __NET_CONFIG_H
#define __NET_CONFIG_H

// includes
#include "types.h"
#include "msgring.h"
#include "spinlock.h"
#include "cache.h"
#include "mips-exts.h"
#include "device.h"

#define PHY_CONTROL                     0x00
#define PHY_STATUS                      0x01
#define PHY_DEVICE_IDENTIFIER1          0x02
#define PHY_DEVICE_IDENTIFIER2          0x03
#define AUTO_NEG_ADVER                  0x04
#define AUTO_NEG_LINK_PARTNER           0x05
#define AUTO_NEG_EXPANSION              0x06
#define AUTO_NEG_TRANSMIT_NXT_PG        0x07
#define AUTO_NEG_LINK_PARTNER_NXT_PG_RECV       0x08
#define BASE1000_T_CTL1                 0x09
#define BASE1000_T_STAT1                0x0A

/* Phy Main Register Address with Reference to the Data Sheet   */
#define BASE1000_T_STAT_EXT1            0x0F
#define BASE1000_T_STAT_EXT             0x10
#define BASE1000_T_STAT_EXT2            0x11
#define BYPASS_CTL                      0x12
#define EXTD_CTL_STAT                   0x16
#define EXT_PHY_CTL_SET1                0x17
#define EXT_PHY_CTL_SET2                0x18
#define INTERRUPT_MASK                  0x19
#define INI_STATUS                      0x1A
#define MAC_AUTO_NEG_CTL_STA            0x1B
#define DEV_AUX_CTL_STAT                0x1C
#define LED_MODE_SELECT                 0x1D
#define LED_BEHAVIOUR                   0x1E

// top config structure

typedef struct net_config_struct {
    int     gmac0; 			// Set 1 if gmac 0 should be initialized
    int     gmac1; 			// Set 1 if gmac 1 should be initialized
    int     gmac2; 			// Set 1 if gmac 2 should be initialized
    int     gmac3; 			// Set 1 if gmac 3 should be initialized
    int     gmac4;                      // Set 1 if gmac 4 should be initialized
    int     gmac5;                      // Set 1 if gmac 5 should be initialized
    int     gmac6;                      // Set 1 if gmac 6 should be initialized
    int     gmac7;                      // Set 1 if gmac 7 should be initialized
    int	    gmac0_xaui;
    int	    gmac4_xaui;

	int		gmac0_frin_spill;
	int		gmac0_frin_spill_entries;
	int		gmac0_frout_spill;
	int		gmac0_frout_spill_entries;
	int		gmac0_rx_spill0;
	int		gmac0_rx_spill0_entries;
	int		gmac0_rx_spill1;
	int		gmac0_rx_spill1_entries;
	int		gmac0_rx_spill2;
	int		gmac0_rx_spill2_entries;
	int		gmac0_rx_spill3;
	int		gmac0_rx_spill3_entries;

	int		gmac1_frin_spill;
	int		gmac1_frin_spill_entries;
	int		gmac1_frout_spill;
	int		gmac1_frout_spill_entries;
	int		gmac1_rx_spill0;
	int		gmac1_rx_spill0_entries;
	int		gmac1_rx_spill1;
	int		gmac1_rx_spill1_entries;
	int		gmac1_rx_spill2;
	int		gmac1_rx_spill2_entries;
	int		gmac1_rx_spill3;
	int		gmac1_rx_spill3_entries;

	int		xgs0_frin_spill;
	int		xgs0_frin_spill_entries;
	int		xgs0_frout_spill;
	int		xgs0_frout_spill_entries;
	int		xgs0_rx_spill0;
	int		xgs0_rx_spill0_entries;
	int		xgs0_rx_spill1;
	int		xgs0_rx_spill1_entries;
	int		xgs0_rx_spill2;
	int		xgs0_rx_spill2_entries;
	int		xgs0_rx_spill3;
	int		xgs0_rx_spill3_entries;

	int		xgs1_frin_spill;
	int		xgs1_frin_spill_entries;
	int		xgs1_frout_spill;
	int		xgs1_frout_spill_entries;
	int		xgs1_rx_spill0;
	int		xgs1_rx_spill0_entries;
	int		xgs1_rx_spill1;
	int		xgs1_rx_spill1_entries;
	int		xgs1_rx_spill2;
	int		xgs1_rx_spill2_entries;
	int		xgs1_rx_spill3;
	int		xgs1_rx_spill3_entries;
	
    int     split_mode; 		// Set 1 if gmacs are in split mode
    int     gmac_rx_enable;             // Set 1 if gmac should set rx_enable as a 
                                        // part of init
    int     gmac_free_descriptors_0;     // Max number of free descriptors that are
                                         // allocated to gmacs. In case of split mode
                                         // this is for context containing gmac0 and
                                         // gmac1.
    int     gmac_free_descriptors_1;     // In case of split mode this is for context 
                                         // containing gmac2 and gmac3.
    uint32_t gmac0_pde_mask_0;		// GMAC0 PDE mask - first 32 bit
    uint32_t gmac0_pde_mask_1;		// GMAC0 PDE mask - second 32 bit
    int     gmac1_free_descriptors_0;    // Max number of free descriptors that are
                                         // allocated to gmacs. In case of split mode
                                         // this is for context containing gmac4 and
                                         // gmac5.
    int     gmac1_free_descriptors_1;    // In case of split mode this is for context
                                         // containing gmac6 and gmac7.
    uint32_t gmac1_pde_mask_0;		// GMAC1 PDE mask  - first 32 bit
    uint32_t gmac1_pde_mask_1;		// GMAC1 PDE mask  - second 32 bit


  int      rgmii_mode;                  // for Condor based silicon, this field
                                        // detenlmnes what type gmac port zero will be
                                        // 0 = std SGMII; 1 = std RGMII

  int fifo_mode0;
  int fifo_mode1;
  int fifo_mode2;
  int fifo_mode3;

    /* 
     * For interface 0 or 1, xgmac and spi4 are exclusive.
     * i.e. for the same interface, both xgmac and spi4 can-not
     * be set. These settings also have to match with board
     * settings.
     * ATX-I board both interfaces are SPI4 on the board
     * ATX-II board both interfaces are XGMAC on the board
     * ATX-III (or the PCI-X daughter card), does not have any of
     * these interfaces.
     */

    int     spi0; 		// Set 1 if spi4 interface 0 should be initialized
    int     xgmac0; 		// Set 1 if xgmac interface 0 should be initialized
    int     xgs0_free_descriptors_0; // Max number of free descriptors for
                                            // 10 gig interface 0
    uint32_t xgs0_pde_mask_0;	//PDE mask for xmgac0 / spi-0 - first 32bit
    uint32_t xgs0_pde_mask_1;	//PDE mask for xmgac0 / spi-0 - second 32bit

    int     spi1; 		// Set 1 if spi4 interface 1 should be initialized
    int     xgmac1; 		// Set 1 if xgmac interface 1 should be initialized
    int     xgs1_free_descriptors_1; // Max number of free descriptors for
                                            // 10 gig interface 1
    uint32_t xgs1_pde_mask_0;	//PDE mask for xgmac1 / spi-1 - first 32 bit
    uint32_t xgs1_pde_mask_1;	//PDE mask for xgmac1 / spi-1 - second 32 bit

    int     abm; 		// Set Automatic memory management by default
    char*   packet_mem_start;  // Virtual Start of Packet memory
    uint64_t    packet_mem_size;       // Size of the Packet memory
    int            gmac0_auto_negotiation; //Set 1 if auto negotiation should be enabaled for gmac0
    int            gmac1_auto_negotiation; //Set 1 if auto negotiation should be enabaled for gmac1
    int            gmac2_auto_negotiation; //Set 1 if auto negotiation should be enabaled for gmac2
    int            gmac3_auto_negotiation; //Set 1 if auto negotiation should be enabaled for gmac3
    int            gmac4_auto_negotiation; //Set 1 if auto negotiation should be enabaled for gmac4
    int            gmac5_auto_negotiation; //Set 1 if auto negotiation should be enabaled for gmac5
    int            gmac6_auto_negotiation; //Set 1 if auto negotiation should be enabaled for gmac6
    int            gmac7_auto_negotiation; //Set 1 if auto negotiation should be enabaled for gmac7
    int            spi4_0_auto_negotiation; //Set 1 if auto negotiation should be enabaled for spi4
    int            spi4_1_auto_negotiation; //Set 1 if auto negotiation should be enabaled for spi4
    int            default_gmac_speed;          // speed of the gmacs, if Auto negotiation is disbled   
    int            default_gmac_duplexity;      // Duplexity of the gmacs, if auto negotiation is disabled
    int            default_spi4_speed;          // speed of the spi4 if Auto negotiation is disbled   
    int            default_spi4_duplexity;      // Duplexity of the spi4, if auto negotiation is disabled
    int		   gmac_byte_offset[NLM_MAX_GMAC];
    int 	   xgs0_byte_offset[MAX_SPI4_PORT];
    int 	   xgs1_byte_offset[MAX_SPI4_PORT];
    int 	   error_id;                    //In case of Tx Abort, desc are sent to this 
						//station instead of FBID given in the descriptors
    int            regular_buffer_size;         // size of the regular buffer in the desc ctrl reg  
    int            gmac_frame_length[NLM_MAX_GMAC];           
    int            xgs_tx_frame_length[NLM_MAX_XGS];        // incase os spi4 it it set on the vitasse register 
    int            xgs_rx_frame_length[NLM_MAX_XGS];        // incase os spi4 it it set on the vitasse register 

} net_config ;

// structure instantiation
// extern net_config    net_conf;

void xlr_net_init(net_config *net_conf);
extern void xlr_net_deinit( net_config *net_conf );

extern net_config net_conf_gmac;
extern net_config net_conf_xgmac;
extern net_config net_conf_spi4;

#endif
