/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/

#ifndef _PHXDESC_H_
#define _PHXDESC_H_

/* Error code is used to indicate any errors */

typedef long phxsec_error_t;

#define PHXSEC_ERR_NONE                    0
#define PHXSEC_ERR_CIPHER_OP              -1
#define PHXSEC_ERR_CIPHER_TYPE            -2
#define PHXSEC_ERR_CIPHER_MODE            -3
#define PHXSEC_ERR_CIPHER_INIT            -4
#define PHXSEC_ERR_DIGEST_TYPE            -5
#define PHXSEC_ERR_DIGEST_INIT            -6
#define PHXSEC_ERR_DIGEST_SRC             -7
#define PHXSEC_ERR_CKSUM_TYPE             -8
#define PHXSEC_ERR_CKSUM_SRC              -9
#define PHXSEC_ERR_ALLOC                  -10
#define PHXSEC_ERR_CONTROL_VECTOR         -11
#define PHXSEC_ERR_LOADHMACKEY_MODE       -12
#define PHXSEC_ERR_PADHASH_MODE           -13
#define PHXSEC_ERR_HASHBYTES_MODE         -14
#define PHXSEC_ERR_NEXT_MODE              -15
#define PHXSEC_ERR_PKT_IV_MODE            -16
#define PHXSEC_ERR_LASTWORD_MODE          -17
#define PHXSEC_ERR_PUBKEY_OP              -18
#define PHXSEC_ERR_SYMKEY_MSGSND          -19
#define PHXSEC_ERR_PUBKEY_MSGSND          -20
#define PHXSEC_ERR_SYMKEY_GETSEM          -21
#define PHXSEC_ERR_PUBKEY_GETSEM          -22

/* this is passed to packet setup to optimize */

#define PHX_SETUP_OP_CIPHER              0x00000001
#define PHX_SETUP_OP_HMAC                0x00000002
#define PHX_SETUP_OP_CIPHER_HMAC         (PHX_SETUP_OP_CIPHER | PHX_SETUP_OP_HMAC)

/* this is passed to control_setup to update w/preserving existing keys */

#define PHX_SETUP_OP_PRESERVE_HMAC_KEY    0x80000000
#define PHX_SETUP_OP_PRESERVE_CIPHER_KEY  0x40000000
#define PHX_SETUP_OP_UPDATE_KEYS          0x00000010
#define PHX_SETUP_OP_FLIP_3DES_KEY        0x00000020

typedef struct tagCipherSetup {
	PHX_CIPHER_MODE   cipher_mode;  
	PHX_CIPHER_TYPE   cipher_type;
	PHX_CIPHER_INIT   cipher_init;
	int               cipher_key_len;
} CipherSetup;

typedef struct tagDigestSetup {
	PHX_DIGEST_TYPE   digest_type;
	PHX_DIGEST_INIT   digest_init;
} DigestSetup;

typedef struct tagChecksumSetup {
	PHX_CKSUM_TYPE    cksum_type;
} ChecksumSetup;

typedef struct tagControlSetup {
	CipherSetup   cipher_setup;
	DigestSetup   digest_setup;
	ChecksumSetup checksum_setup;
	u8            crypt_key[PHXSEC_MAX_CRYPT_KEY_LENGTH];
	u8            mac_key[PHXSEC_MAX_BLOCK_SIZE];

} ControlSetup;

/* this structure is used during packet processing: 
 * no checksum setups, no loading keys (maybe just 
 * flipping in place)
 */

typedef struct tagShortControlSetup {
	CipherSetup   cipher_setup;
	DigestSetup   digest_setup;
} ShortControlSetup;

typedef struct tagPacketSetup {

	PHX_CIPHER_OP         cipher_op;
	PHX_DIGEST_SRC        digest_src;
	PHX_CKSUM_SRC         cksum_src;
	u32                   nonce;
	u32                   cfb_mask;
	u32                   cipher_offset;
	u32                   digest_offset;
	u32                   cksum_offset;
	PHX_LOADHMACKEY_MODE  pkt_hmac;
	PHX_PADHASH_MODE      pkt_hash;
	PHX_HASHBYTES_MODE    pkt_hashbytes;
	PHX_NEXT_MODE         pkt_next;
	PHX_PKT_IV_MODE       pkt_iv;
	PHX_LASTWORD_MODE     pkt_lastword;
	u8                   *source_buf;
	u32                   source_buf_size;
	u8                   *dest_buf;
	u32                   dest_buf_size;
	u8                   *auth_dest;
	u8                   *cksum_dest;
	u32                  iv_offset;
} PacketSetup;

#endif /* _PHXDESC_H_ */
