/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/
#ifndef __PHYSADDRSPACE_H__
#define __PHYSADDRSPACE_H__

#include "system.h"

#define KB(x) ((x) << 10)
#define MB(x) ((x) << 20)
#define GB(x) ((x) << 30)
#define TB(x) ((x) << 40)

#define PSB_MEM_MAP_MAX 32
#define PSB_IO_MAP_MAX 32

/* The Processor address space (1TB) is divided into: */
enum {
	PHYS_MEMORY = 1,
	PHYS_IO,
	EXCEPTION_VECTORS,
	BOOTLOADER_0,
	NMI_MEMORY,
	PHYS_AVAILABLE
};

/* The PHYS_MEMORY, some areas are RAM and some areas are ROM */
enum {
	BOOT_MEM_RAM      =  1,
	BOOT_MEM_ROM_DATA =  2,
	BOOT_MEM_RESERVED =  3
};

/* There could be a number of them */
struct psb_mem_map {
	int nr_map;
	struct psb_mem_map_entry {
		uint64_t addr;	/* start of memory segment */
		uint64_t size;	/* size of memory segment */
		long type;	/* type of memory segment */
	} map[PSB_MEM_MAP_MAX];
};

/* The PHYS_IO is divided into: */
enum {  
	PROCESSOR_PIO_SPACE = 0x10, /* The Processor's Peripheral 
				       and IO Configuration Space */
        PCIX_IO_SPACE,
        PCIX_CFG_SPACE,
        PCIX_MEMORY_SPACE,
        HT_IO_SPACE,
        HT_CFG_SPACE,
        HT_MEMORY_SPACE,
        SRAM_SPACE,
        FLASH_CONTROLLER_SPACE,
        PCIE_IO_SPACE,
        PCIE_CFG_SPACE,
        PCIE_MEMORY_SPACE
};

/*
  Each of the above is a separate region. For e.g.
  PROCESSOR_PIO_SPACE starts typically at 0x1ef00000
  and is of size 1MB 
*/
struct psb_io_map {
	int nr_map;
	struct psb_io_map_entry {
		uint64_t addr;	/* start of IO segment */
		uint64_t size;	/* size of IO segment */
		long type;	/* type of IO segment */
	} map[PSB_IO_MAP_MAX];
};

/* 
   Use bits 16-32 to indicate whether an area is shared,
   can be used for storing env variables, etc.
   Bits 15-0 could be used to indicate the resource type
   PHYS_MEMORY,PHYS_IO etc.
*/
enum region_attrib{
	SHARED    = 0x00010000,
	BOOTCODE  = 0x00020000,
	BKUPCODE  = 0x00040000,
	ENVSTORE  = 0x00080000,
};

struct phys_region {
	const char *name;
	int type;
	int busy;
	uint64_t base;
	uint64_t size;
	struct phys_region *next;
};

extern uint64_t physaddr_reserve_region(const char *name, int type,
				 uint64_t base, uint64_t size);

void physaddr_release_region(const char *name, int type,uint64_t base, uint64_t size);
#endif
