/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/
#ifndef __SMP_H__
#define __SMP_H__

#include "types.h"
#include "spinlock.h"
#include "xlr_boot_lib.h"

#ifndef __ASSEMBLY__


static __inline__ void smp_lock_init(spinlock_t * lp)
{
	__asm__ __volatile__(".set noreorder    \n"
			     "    sync          \n"
			     "    sw $0, %0     \n"
			     ".set\treorder     \n"
			     :"=m"(*lp)
			     :"m"(*lp)
			     :"memory");
}

static __inline__ void smp_lock(spinlock_t * lp)
{
	unsigned int tmp;

	__asm__ __volatile__(".set noreorder    \n"
			     "1: ll    %1, %2   \n"
			     "   bnez  %1, 1b   \n"
			     "   nop            \n"
			     "   li    %1, 1    \n"
			     "   sc    %1, %0   \n"
			     "   beqz  %1, 1b   \n"
			     "   nop            \n"
			     "   sync           \n"
			     ".set reorder      \n"
			     :"=m"(*lp), "=&r"(tmp)
			     :"m"(*lp)
			     :"memory");
}
#define smp_unlock(x) smp_lock_init(x)

extern uint32_t num_cpus_online;
extern uint32_t cpu_online_map;

extern void smp_init(struct processor *this_processor);
void smp_call_function(void *fn, void *args, uint32_t cpu_mask);
void setup_per_cpu_areas(void);
void wakeup_cpu_msgring(int cpu, unsigned long fn, unsigned long args);

extern void xlr_park(unsigned long);

#endif

#endif
