/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/

#ifndef _SPI4_H
#define _SPI4_H

#include <netdevice.h>
//#include <mac.h>
#include "device.h"


// register defs
#define CNTRL_REG          0x6C

#define SPI_TX_CAL_LEN     0x0
#define SPI_TX_CAL_M       0x4
#define SPI_RX_CAL_LEN     0x18    
#define SPI_RX_CAL_M       0x1c

#define SPI_INT            0x70
#define SPI_INT_MASK       0x74
#define SPI_TX_STATUS      0x7C
#define SPI_RX_STATUS      0x80

#define EGR_MAXBURST1      0x30
#define EGR_MAXBURST2      0x34
#define ING_MAXBURST1      0x38
#define ING_MAXBURST2      0x3C
#define SPI_BE_BURST       0x78

#define SPI_RX_FIFO_BASE   0x90
#define SPI_RX_FIFO_SIZE   0x88
#define SPI_TX_FIFO_BASE   0x94
#define SPI_TX_FIFO_SIZE   0x8C

#define GLU_RX_FIFO_BASE   0x229
#define GLU_TX_FIFO_BASE   0x221
#define SPI_ALIGN_0        0x240
#define SPI_ALIGN_1        0x241
#define SPI_ALIGN_2        0x242
#define SPI_ALIGN_3        0x243

#define SPI_CENTER_TAP     0x44444444

#define XLR_TOTAL_SPI4_CHANNELS 0xa

/********** CPLD register     **************/
#define CPLD_MISC_STATUS_REG  0x0e
#define SPI4_MASK_BIT1 1



void spi4_arizona_setup(int num);
void spi4_inject_free_desc(int num_desc, int desc_bucket);

int spi4_tx(int port);
int spi4_rx(int bucket);
int spi4_0_init();
int spi4_1_init();
void spi4_pde_setup(unsigned int cpu_mask, unsigned int app_mask, unsigned int *mmio);
void spi4_setup_pde_mask(unsigned int *, uint32_t , uint32_t );
int  spi4_card_detect(int device_number);
void spi4_rx_fifo_carve(int num_interface, int num_ports);
void spi4_tx_fifo_carve(int num_interface, int num_ports);
void spi4_set_rx_maxburst1(int num_interface, int value);
void spi4_set_rx_maxburst2(int num_interface, int value);
void spi4_set_tx_maxburst1(int num_interface, int value);
void spi4_set_tx_maxburst2(int num_interface, int value);
void spi4_set_static_alignment(int num_interface, int value);
int  check_spi4_link_status();
int  enable_spi4_autonegotiation(int port );
int  disable_spi4_autonegotiation(int port, int speed, int duplexity );
void get_spi4_speed_duplexity(int spi4);
void  change_vits_base(int port);
int set_enable_auto_negotiation( int *spi4_enable_auto_nego, int port );

int nlm_spi4_init( device_conf *current_dev);
int nlm_spi4_port_open( device_conf * current_dev, int port );
int nlm_spi4_port_close( device_conf * current_dev, int port);
int nlm_check_spi4_speed(device_conf * current_dev, int port);


/*
 * 
 */
#define SET_INDEX_FIELD(mmio,addr,index,value) \
  phoenix_write_reg(mmio, addr, ((value) | ((index) << 16)))

/***************************************************
 * SPI_RESET_CONFIG (Address Offset 0x6C) SPI4 Interface Config
 **************************************************/

/* #define MORE_TRN_EN  ; 
 */
#define MORE_TRN_EN_LSB     14
#define MORE_TRN_EN_BITS    ONE_BIT
#define MORE_TRN_EN_MASK    (MORE_TRN_EN_BITS << MORE_TRN_EN_LSB)

/* #define TX_FRM_ERR_EN  ; 
 */
#define TX_FRM_ERR_EN_LSB     13
#define TX_FRM_ERR_EN_BITS    ONE_BIT
#define TX_FRM_ERR_EN_MASK    (TX_FRM_ERR_EN_BITS << TX_FRM_ERR_EN_LSB)

/* #define RX_TRN_LOS  ; 
 */
#define RX_TRN_LOS_LSB     12
#define RX_TRN_LOS_BITS    ONE_BIT
#define RX_TRN_LOS_MASK    (RX_TRN_LOS_BITS << RX_TRN_LOS_LSB)

/* #define RX_TRN_RESET  ; 
 */
#define RX_TRN_RESET_LSB     11
#define RX_TRN_RESET_BITS    ONE_BIT
#define RX_TRN_RESET_MASK    (RX_TRN_RESET_BITS << RX_TRN_RESET_LSB)

/* #define FIFO_SW_RESET  ; 
 */
#define FIFO_SW_RESET_LSB     10
#define FIFO_SW_RESET_BITS    ONE_BIT
#define FIFO_SW_RESET_MASK    (FIFO_SW_RESET_BITS << FIFO_SW_RESET_LSB)

/* #define DDL_ENABLE  ; 
 */
#define DDL_ENABLE_LSB     9
#define DDL_ENABLE_BITS    ONE_BIT
#define DDL_ENABLE_MASK    (DDL_ENABLE_BITS << DDL_ENABLE_LSB)

/* #define SPI_DEP_LPB  ; 
 */
#define SPI_DEP_LPB_LSB     8
#define SPI_DEP_LPB_BITS    ONE_BIT
#define SPI_DEP_LPB_MASK    (SPI_DEP_LPB_BITS << SPI_DEP_LPB_LSB)

/* #define SPI_SHAL_LPB  ; 
 */
#define SPI_SHAL_LPB_LSB     7
#define SPI_SHAL_LPB_BITS    ONE_BIT
#define SPI_SHAL_LPB_MASK    (SPI_SHAL_LPB_BITS << SPI_SHAL_LPB_LSB)

/* #define USR_DEEP_LPB  ; 
 */
#define USR_DEEP_LPB_LSB     6
#define USR_DEEP_LPB_BITS    ONE_BIT
#define USR_DEEP_LPB_MASK    (USR_DEEP_LPB_BITS << USR_DEEP_LPB_LSB)

/* #define USR_SHAL_LPB  ; 
 */
#define USR_SHAL_LPB_LSB     5
#define USR_SHAL_LPB_BITS    ONE_BIT
#define USR_SHAL_LPB_MASK    (USR_SHAL_LPB_BITS << USR_SHAL_LPB_LSB)

/* #define SPI4_RX_ENABLE  ; 
 */
#define SPI4_RX_ENABLE_LSB     3
#define SPI4_RX_ENABLE_BITS    ONE_BIT
#define SPI4_RX_ENABLE_MASK    (SPI4_RX_ENABLE_BITS << SPI4_RX_ENABLE_LSB)

/* #define SPI4_RX_SW_RESET  ; 
 */
#define SPI4_RX_SW_RESET_LSB     2
#define SPI4_RX_SW_RESET_BITS    ONE_BIT
#define SPI4_RX_SW_RESET_MASK    (SPI4_RX_SW_RESET_BITS << SPI4_RX_SW_RESET_LSB)

/* #define SPI4_TX_ENABLE  ; 
 */
#define SPI4_TX_ENABLE_LSB     1
#define SPI4_TX_ENABLE_BITS    ONE_BIT
#define SPI4_TX_ENABLE_MASK    (SPI4_TX_ENABLE_BITS << SPI4_TX_ENABLE_LSB)

/* #define SPI4_TX_SW_RESET  ; 
 */
#define SPI4_TX_SW_RESET_LSB     0
#define SPI4_TX_SW_RESET_BITS    ONE_BIT
#define SPI4_TX_SW_RESET_MASK    (SPI4_TX_SW_RESET_BITS << SPI4_TX_SW_RESET_LSB)

/*************************************************************************
 * SPI_INT (Address Offset 0x70) SPI4 Interface Interrupt Status
 * SPI_INT_MASK (Address Offset 0x74) SPI4 Interface Interrupt Status Mask
 ************************************************************************/

/* #define INVALID_CW  ; 
 */
#define INVALID_CW_LSB     3
#define INVALID_CW_BITS    ONE_BIT
#define INVALID_CW_MASK    (INVALID_CW_BITS << INVALID_CW_LSB)

/* #define DIP2_ERROR  ; 
 */
#define DIP2_ERROR_LSB     2
#define DIP2_ERROR_BITS    ONE_BIT
#define DIP2_ERROR_MASK    (DIP2_ERROR_BITS << DIP2_ERROR_LSB)

/* #define DIP4_ERROR  ; 
 */
#define DIP4_ERROR_LSB     1
#define DIP4_ERROR_BITS    ONE_BIT
#define DIP4_ERROR_MASK    (DIP4_ERROR_BITS << DIP4_ERROR_LSB)

/*************************************************************************
 * SPI_TX_STATUS (Address Offset 0x7C) SPI4 Interface Tx Status
 ************************************************************************/

/* #define TX_SYNC  ; 
 */
#define TX_SYNC_LSB     2
#define TX_SYNC_BITS    ONE_BIT
#define TX_SYNC_MASK    (TX_SYNC_BITS << TX_SYNC_LSB)

/* #define TX_TRAIN_DETECT  ; 
 */
#define TX_TRAIN_DETECT_LSB     1
#define TX_TRAIN_DETECT_BITS    ONE_BIT
#define TX_TRAIN_DETECT_MASK    (TX_TRAIN_DETECT_BITS << TX_TRAIN_DETECT_LSB)

/* #define TX_CAL_ERROR  ; 
 */
#define TX_CAL_ERROR_LSB     0
#define TX_CAL_ERROR_BITS    ONE_BIT
#define TX_CAL_ERROR_MASK    (TX_CAL_ERROR_BITS << TX_CAL_ERROR_LSB)

/*************************************************************************
 * SPI_RX_STATUS (Address Offset 0x80) SPI4 Interface Rx Status
 ************************************************************************/

/* #define MISSING_EOP  ; 
 */
#define MISSING_EOP_LSB     4
#define MISSING_EOP_BITS    ONE_BIT
#define MISSING_EOP_MASK    (MISSING_EOP_BITS << MISSING_EOP_LSB)

/* #define MISSING_SOP  ; 
 */
#define MISSING_SOP_LSB     3
#define MISSING_SOP_BITS    ONE_BIT
#define MISSING_SOP_MASK    (MISSING_SOP_BITS << MISSING_SOP_LSB)

/* #define RX_SYNC  ; 
 */
#define RX_SYNC_LSB     1
#define RX_SYNC_BITS    ONE_BIT
#define RX_SYNC_MASK    (RX_SYNC_BITS << RX_SYNC_LSB)

/* #define RX_TRAIN_DETECT  ; 
 */
#define RX_TRAIN_DETECT_LSB     0
#define RX_TRAIN_DETECT_BITS    ONE_BIT
#define RX_TRAIN_DETECT_MASK    (RX_TRAIN_DETECT_BITS << RX_TRAIN_DETECT_LSB)


#endif


