/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/

#ifndef _UART_H
#define _UART_H

#include <pt_regs.h>

#define UART_RHR 0
#define UART_THR 0
#define UART_IER 1
#define UART_IIR 2
#define UART_FCR 2
#define UART_LCR 3
#define UART_MCR 4
#define UART_LSR 5
#define UART_MSR 6

#define UART_DLB_1 0
#define UART_DLB_2 1

#define UART_DEBUG_1 8
#define UART_DEBUG_2 9

#ifdef XLP_SIM
// baud rate divisors
#define UART_BR9600_DLB1 0x61
#define UART_BR9600_DLB2 0x03

#define UART_BR38400_DLB1 0xd8
#define UART_BR38400_DLB2 0x00

#define UART_BR115200_DLB1 0x48
#define UART_BR115200_DLB2 0x00

#else
// baud rate divisors
#define UART_BR9600_DLB1 0xad
#define UART_BR9600_DLB2 0x01

#define UART_BR38400_DLB1 0x6b
#define UART_BR38400_DLB2 0x00

#define UART_BR115200_DLB1 0x23
#define UART_BR115200_DLB2 0x00
#endif

#define UART_BR_DLB1 UART_BR115200_DLB1
#define UART_BR_DLB2 UART_BR115200_DLB2

#ifndef __ASSEMBLY__

#define TTYS0 0
#define TTYS1 1
#define VIRT_UART 2
#define VIRT_UART_OPENED 0xeadbeef
#define USER_CMD_SIZE    (1*1024)
#define USER_RESULT_SIZE (7*1024)
#define VIRTUAL_UART_NR   32
#ifdef _ABI64
	#define VIRT_UART_BUF_START ((48<<20) | 0xffffffff80000000ULL)
#else
	#define VIRT_UART_BUF_START ((48<<20) | 0x80000000)
#endif

typedef struct test{
        volatile unsigned char *rx_fifo;
        volatile unsigned int *rx_pro;
        volatile unsigned int *rx_con;
        volatile unsigned char *tx_fifo;
        volatile unsigned int *tx_pro;
        volatile unsigned int *tx_con;
        volatile int *status;
}virt_uart_struct;

extern void outbyte(char c);
extern int inbyte(void);
extern void uart_int_handler(int irq, struct pt_regs *regs);
extern int inbyte_nowait(void);
#endif

#endif
