/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/
#include <sys_entry.h>
#include <stdio.h>
#include <semlib.h>
#include <barrier_lib.h>
#define dbg(a,b...) //printf("\n[LOADER--> %s @ %d] "a"\n",__FUNCTION__,__LINE__, ##b)

/*This is a call back from netos library to netos loader.
  Various system calls - shmget, semaphores, barriers
  needs this.
 */
int nlm_sys_call(int index, void *arg)
{
    int ret = 0;
    dbg("index %d, arg %p enter",index, arg);
    switch(index)
    {
        case NLM_SHM_SYS_CALL:
	    ret = shm_sys_call(arg);
            break;
        case NLM_SEM_SYS_CALL:
            dbg("Calling sem_syscall..");
            ret = sem_sys_call(arg);
            break;
        case NLM_BARRIER_SYS_CALL:
            dbg("Calling barrier_syscall..");
            ret = barrier_sys_call(arg);
            break;
	case NLM_CONTIGALLOC_SYS_CALL:
	    ret = contig_alloc_sys_call(arg);
	    break;			
	case NLM_IRQ_SYS_CALL:
	    ret = config_irq_sys_call(arg);
	    break;
	case NLM_EXCEPT_SYS_CALL:
	    ret = config_exception_sys_call(arg);
	    break;
        default:
            ret = -1;
            break;
    }

    dbg("index %d, arg %p exit",index, arg);
    return ret;
}

void system_call_init()
{
	/*Init semaphore data structures..*/
	semaphore_init();

	/* Init shared mem data structures */
	netl_shm_init();

	/*Init barrier data structures..*/
	barrier_init();
}
