/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/
/*This is the main bootloader lib file.*/
#include <string.h>

#include <stackframe.h>
#include "cpu.h"
#include "rmidelay.h"
#include "xlr_boot_lib.h"
#include "mm.h"
#include "shared_structs.h"
#include "smp.h"
#include "physaddrspace.h"
#ifdef XLP_SIM
#include "nlm_hal_pic.h"
#else
#include "pic.h"
#endif
#include "system.h"
#include "bitops.h"

static struct boot1_info *xlr_lib_boot1_info=NULL;
static unsigned long kseg_prelaunch_init_args = 0;
static unsigned char loader_lock[16] = {0};
static unsigned long global_wakeup_mask = 0;
volatile struct xlr_vcpu_wakeup_info *xlr_vcpu_wakeup_info = NULL;

void xlr_park(unsigned long args);
static void xlr_lib_kuseg_helper(void);
static void load_scratch_info(void);
static void disable_ie(void);
static void enable_ie(void);

static volatile int xlr_wakeup_ipi[XLR_NUM_CPU] = {0};
static int loading_kseg0 = 0;
static uint64_t released_phys_addr[32];
static uint64_t released_phys_size[32];
static uint64_t released_phys_entry;

extern void flush_tlb_all (void);
extern void per_cpu_init (void);
extern int printf(const char *fmt, ...);
extern void xlr_lib_prelaunch_init_kseg(unsigned long);

extern struct thread_info threads_info[MAX_CPUS] __section(".thread.info");
extern struct xlr_vcpu_wakeup_info xlr_wakeup_info[32];
extern int init_irq(void);

#ifndef __STR
#define __STR(x) #x
#endif
#ifndef STR
#define STR(x) __STR(x)
#endif

static __inline__ void lib_sync()
{
	__asm__ volatile(
		"sync\n"
		);
}

static __inline__ void load_scratch_info(void)
{
	__asm__ volatile(
		".set push\n"
		".set noreorder\n"
		".set mips64\n"
		"dmtc0 %0, $22, 0\n"	/* move (&boot1_info) into $22    */
		"dmtc0 %1, $22, 1\n"	/* move (master cpu + buddy mask) into $22 */
		"dmtc0 %2, $22, 7\n"
		".set pop\n"
//		:: "r"(xlr_lib_boot1_info),"r"(xlr_lib_boot1_info->userapp_cpu_map),
		:: "r"(xlr_lib_boot1_info),"r"(xlr_lib_boot1_info),
		"r"(xlr_vcpu_wakeup_info)
		);
}

static __inline__ void disable_ie()
{
	__asm__ volatile (".set push     \n"
			  ".set noreorder        \n"
			  ".set noat             \n"
			  "mfc0 $1, $12          \n"
			  "ori  $1, 0x1          \n"
			  "xori $1, 0x1          \n"
			  "mtc0 $1, $12          \n"
			  "nop                   \n"
			  "nop                   \n"
			  "nop                   \n"
			  "nop                   \n"
			  ".set pop              \n"
		);
}

static __inline__ void enable_ie()
{
	__asm__ volatile (".set push             \n"
			  ".set noreorder        \n"
			  ".set noat             \n"
			  "mfc0 $1, $12          \n"
			  "ori  $1, 0x1f         \n"
			  "xori $1, 0x1e         \n"
			  "mtc0 $1, $12          \n"
			  "nop                   \n"
			  "nop                   \n"
			  "nop                   \n"
			  "nop                   \n"
			  ".set pop              \n"
		);
}

void xlr_lib_reclaim_resources(void)
{
	struct xlr_vcpu_wakeup_info *t = (struct xlr_vcpu_wakeup_info *)
		xlr_vcpu_wakeup_info + processor_id();

	t->cpu_status = CPU_WAIT_TO_RELEASE_RESOURCE;
	lib_sync();

	while(t->cpu_status != CPU_RESOURCE_RELEASED);

	t->cpu_status = CPU_STOPPED;
	lib_sync();
}

void xlr_lib_entry(unsigned long args)
{
	struct xlr_vcpu_wakeup_info *t = (struct xlr_vcpu_wakeup_info *)
		xlr_vcpu_wakeup_info + processor_id();

	extern void xlr_lib_entry_asm(unsigned long long, unsigned long long,
			void *, unsigned long long);

	xlr_lib_entry_asm(t->sp, t->gp, (void *)(long)xlr_park, 0);
}

void set_argv_envs(struct xlr_vcpu_wakeup_info *t)
{
	int i=0;
	unsigned char *tmp=NULL;
	tmp = t->argv_buf;

	for(i=0;i<t->argc;i++){
		t->argv[i] = tmp;
		while(*tmp != ' ' && *tmp)
			tmp++;
		tmp++;
	}

	for(; i<XLR_MAX_ARGV; i++)
		t->argv[i] = NULL;

	tmp = t->env_buf;
	for(i=0;i<t->env;i++){
		t->envs[i] = tmp;
		while(*tmp != ' ' && *tmp)
			tmp++;
		tmp++;
	}
}

void xlr_start_app(void)
{
	int vcpu = processor_id();

	struct xlr_vcpu_wakeup_info *t = (struct xlr_vcpu_wakeup_info *)
		(xlr_vcpu_wakeup_info + vcpu);

	unsigned long long arg1 = 0;
	unsigned long long arg2 = 0;
	unsigned long long arg3 = 0;
	unsigned long long arg4 = 0;

	extern void call_start_func_asm(unsigned long,
					unsigned long, unsigned long,
					unsigned long, unsigned long,
					unsigned long, unsigned long,
					unsigned long);

	t->cpu_status = CPU_RUNNING;

	lib_spin_lock(((lib_spinlock_t *)t->loader_lock));
	*(t->global_wakeup_mask) = global_wakeup_mask & ~(1UL<<vcpu);
	lib_spin_unlock(((lib_spinlock_t *)t->loader_lock));

	flush_tlb_all();

	/*Reset STATUS Register*/
#if defined(XLP_SIM)
	write_c0_status(read_c0_status() | 0x70000000);
#else
	write_c0_status(read_c0_status() | 0x50000000);
#endif

	/* Call Helper And Setup TLB For KUSEG Apps */
	if (t->mode == KUSEG_APP){
		xlr_lib_kuseg_helper();
	}

	/* Set argv and envs with proper pointers */
	set_argv_envs(t);

	ldr_flush_icache_all();
	ldr_flush_dcache_all();

	/*
	 * Setup $22,0 with boot1_info structure. 
	 * This may required for initialization process.		
	 * Setup $22,7 with "xlr_vcpu_wakeup_info" base addr
	 * Rmios lib startup routine will find the 
	 * appropriate offset based on this.
	 */
	load_scratch_info();

#ifdef DEAD_CODE
	/* NETOS_TODO: Remove this. This has to come from FDT */

	if(!loading_kseg0)
		rmi_udelay(10000000);
#endif

	/* Case where we are launching kseg application  
	 * on vcpu not running loader, Setup all the 
	 * required arguments accordingly
	 */
	if(t->kseg_master == 1) {
		arg1 = (long)t->argc;
		arg2 = (long)t->argv;
		arg3 = (long)t->envs;
		arg4 = (long)xlr_lib_boot1_info;
		call_start_func_asm(arg1, arg2, arg3, arg4,
				(unsigned long)t->func, t->gp, t->sp,
				(unsigned long)t->reentry_function);
	}
	else {
		arg1 = (long)fw_arg0;
		arg2 = (long)fw_arg1;
		arg3 = (long)fw_arg2;
		arg4 = (long)t->args; /* User passed arg throug clone syscall */
/* NETOS_HACK: should fix so call_start_func_asm works */
#if 1
		__asm__ __volatile__(
			".set push\n"
			".set noreorder\n"
			"move $4,  %0\n"
			"move $5,  %1\n"
			"move $6, %2\n"
			"move $7, %3\n"
			"move $25, %4\n"
			"move $28, %5\n"
			"move $31, %6\n"
			"move $29, %7\n"
			"jr $25\n"
			"nop\n"
			".set pop\n"
			:
			: "r"(arg1), "r"(arg2), "r"(arg3), "r"(arg4), 
			  "r"(t->func), "r"(t->gp), "r"(t->reentry_function), "r"(t->sp) 
			: "$4", "$5", "$6", "$7", "$25", "$31", "$29"
			);
#else
		call_start_func_asm(arg1, arg2, arg3, arg4, 
							(unsigned long) t->func, t->gp, t->sp, 
							(unsigned long) t->reentry_function);
#endif
	}
}

void xlr_send_stop_ipi(unsigned long mask)
{
#ifdef NETOS_LOADER_TODO
	int ipi;
	int pid, tid;
	xlr_lib_phnx_reg_t *mmio = phoenix_io_mmio(PIC);
	volatile struct xlr_vcpu_wakeup_info *m = xlr_vcpu_wakeup_info;
	volatile struct xlr_vcpu_wakeup_info *t = NULL;
	volatile int i;

	lib_spin_lock(((lib_spinlock_t *)m->loader_lock));

	/* Send stop ipi to all requested vcpus
	 */
	for(i=0; i<XLR_NUM_CPU; i++){
		if(mask & (1U<<i)){
#if defined(XLP_SIM)
			ipi = tid = pid = mmio = 0; /* For compiler sake */
			/* XLP TODO: currently setting the nmi & node id to zero.
			   To be correctly determined by a proper macro */
			nlm_hal_pic_send_ipi(0, IRQ_STOP_CPU_IPI, 0, i);
#else

			pid = i >> 2;
			tid = i % 4;
			ipi = (pid << 20) | (tid << 16) | IRQ_STOP_CPU_IPI ;
			xlr_lib_phnx_write_reg(mmio, PIC_IPI, ipi);
#endif
		}
	}

	/* wait for cpus to come back to park mode
	 */
	rmi_udelay(100000000UL);

	for(i=0; i<XLR_NUM_CPU; i++){
		if(mask & (1U<<i)){
			t = xlr_vcpu_wakeup_info + i;
			if(t->cpu_status == CPU_STOPPED){
#if defined(XLP_SIM)
		                ipi = tid = pid = mmio = 0; /* For compiler sake */
			        /* XLP TODO: currently setting the nmi & node id to zero.
				   To be correctly determined by a proper macro */
				nlm_hal_pic_send_ipi(0, IRQ_RECLAIM_CPU_RSRC_IPI, 0, i);
#else
				pid = i >> 2;
				tid = i % 4;
				ipi = (pid << 20) | (tid << 16) | IRQ_RECLAIM_CPU_RSRC_IPI;
				xlr_lib_phnx_write_reg(mmio, PIC_IPI, ipi);
#endif
				lib_sync();
				/*wait till cpu comes to wait_to_release state*/
				while(t->cpu_status != CPU_WAIT_TO_RELEASE_RESOURCE);
			}
		}
	}

	lib_spin_unlock(((lib_spinlock_t *)m->loader_lock));
#endif /* NETOS_LOADER_TODO */
}

void loader_wakeup_cpu(unsigned int cpu, unsigned long fn, unsigned long args)
{
        volatile struct xlr_vcpu_wakeup_info *t;
        int ipi;
        int pid = cpu >> 2;
        int tid = cpu % 4;

#ifndef XLP_SIM
        phoenix_reg_t *mmio = phoenix_io_mmio(PIC);
#endif
        t = xlr_vcpu_wakeup_info + cpu;
        lib_spin_lock(((lib_spinlock_t *)t->loader_lock));
        t = xlr_vcpu_wakeup_info + cpu;
        t->func = fn;
        t->args = args;

        ipi = (pid << 20) | (tid << 16) | IRQ_WAKEUP_CPU_IPI ;

        *(t->global_wakeup_mask) = *(t->global_wakeup_mask) | (1UL<<cpu);

	/* memory fence for any pending wakeup info stores */
	lib_sync();

#if defined(XLP_SIM)
	/* XLP TODO: currently setting the nmi & node id to zero.
	   To be correctly determined by a proper macro */
		nlm_hal_pic_send_ipi(0, IRQ_WAKEUP_CPU_IPI, 0, cpu);
#else
        xlr_lib_phnx_write_reg(mmio, PIC_IPI, ipi);
#endif

        lib_spin_unlock(((lib_spinlock_t *)t->loader_lock));
}

/* 
 * This function is specific for loader .
 */
static void loader_setup_tlb(unsigned long long virt, unsigned long long phys, int size)
{
    unsigned long  value = 0;
    uint64_t attr = (3 << 3) | (1 << 2) | (1 << 1) | (1 << 0);
    int wired = 0;
    int page_size = tlb_size_to_page_size(size);
    int page_mask = tlb_size_to_mask(page_size);

    write_c0_pagemask(page_mask);
    write_c0_entryhi(virt);

    value = (((phys & 0xffffffffffULL) >> 12) << 6) | attr; 
    write_c0_entrylo0(value);

    value = ((((phys + page_size) & 0xffffffffffUL) >> 12) << 6) | attr;
    write_c0_entrylo1(value);

    wired = read_c0_wired();

    write_c0_index(wired); 
    tlb_write_indexed();

    write_c0_wired(wired + 1);
}

static void xlr_lib_kuseg_helper(void)
{
	struct xlr_vcpu_wakeup_info *t = (struct xlr_vcpu_wakeup_info *)
		xlr_vcpu_wakeup_info + processor_id();
	int i;
	extern unsigned long long app_endaddr;

	write_c0_wired(0);

	if (t->map_count <= 0) {
		printf("(NetOS) cpu-%d: No TLB mappings (Halting CPU!)?\n", processor_id());
		cpu_halt();
	}

	for (i = 0; i < t->map_count; i++){
		loader_setup_tlb((unsigned long long)t->map[i].virt |
			  ((unsigned long long)t->map[i].asid << 32),
			  t->map[i].phys, t->map[i].page_size);
	}

	pte_init(processor_id(), app_endaddr);
}

void xlr_launch_init(void)
{
	int i=0;
	struct xlr_vcpu_wakeup_info *t = xlr_wakeup_info;

	xlr_vcpu_wakeup_info = xlr_wakeup_info;
	memset(loader_lock,0,sizeof(t->loader_lock));

	for(i=0;i<XLR_NUM_CPU;i++){
		t->sp = PTR2U64((char *)&threads_info[i] + THREAD_SIZE - 32);
		t->gp = PTR2U64(&threads_info[i]);
		t->global_wakeup_mask = &global_wakeup_mask;
		t->loader_lock = (lib_spinlock_t*)loader_lock;
		t++;
	}
}

void xlr_lib_release_memory(uint64_t phys, uint64_t size)
{
	released_phys_addr[released_phys_entry] = phys;
	released_phys_size[released_phys_entry] = size;
	released_phys_entry++;
}

static int is_mergable(struct psb_mem_map *tmp,int k)
{
	if((tmp->map[k-1].addr + tmp->map[k-1].size) == tmp->map[k].addr)
		return 1;
	return 0;
}

static void update_memory_map(void)
{
	struct psb_mem_map *avail_mem_map;
	struct psb_mem_map tmp;
	int i=0,j=0,k=0;

	avail_mem_map = (struct psb_mem_map *)(unsigned long)
		xlr_lib_boot1_info->avail_memory_map;
	tmp.nr_map = 0;

	for(i=0;i<avail_mem_map->nr_map && j<released_phys_entry;){
		if(avail_mem_map->map[i].addr > released_phys_addr[j]){
			tmp.map[k].addr = released_phys_addr[j];
			tmp.map[k].size= released_phys_size[j];
			j++;
		}else{
			tmp.map[k].addr = avail_mem_map->map[i].addr;
			tmp.map[k].size= avail_mem_map->map[i].size;
			i++;
		}
		if(k>1 && is_mergable(&tmp,k)){
			tmp.map[k-1].size += tmp.map[k].size;
		}else{
			tmp.map[k].type = BOOT_MEM_RAM;
			k++;
			tmp.nr_map++;
		}
	}
	if(i != avail_mem_map->nr_map){
		while(i != avail_mem_map->nr_map){
			tmp.map[k].addr = avail_mem_map->map[i].addr;
			tmp.map[k].size = avail_mem_map->map[i].size;
			tmp.map[k].type = BOOT_MEM_RAM;
			if(k>1 && is_mergable(&tmp,k)){
				tmp.map[k-1].size += tmp.map[k].size;
			}else{
				k++;
				tmp.nr_map++;
			}
			i++;
		}
	}else{
		while(j != released_phys_entry){
			tmp.map[k].addr = released_phys_addr[j];
			tmp.map[k].size = released_phys_size[j];
			tmp.map[k].type = BOOT_MEM_RAM;
			if(k>1 && is_mergable(&tmp,k)){
				tmp.map[k-1].size += tmp.map[k].size;
			}else{
				k++;
				tmp.nr_map++;
			}
			j++;
		}
	}
	*avail_mem_map = tmp;
}

static void dump_xlr_vcpu_wakeup_info(void)
{
	int i = 0;
	struct xlr_vcpu_wakeup_info *t = 0;

	for(i = 0; i < 1; i++) {

		t = (struct xlr_vcpu_wakeup_info *)xlr_vcpu_wakeup_info + i;
		if (!t->valid) continue;

		printf("============ cpu_%d ============== \n", i);
		printf("func = 0x%p, args = 0x%p\n", (void *) t->func, (void *) t->args);
		printf("sp = 0x%016llx, gp = 0x%016llx\n", t->sp, t->gp);
		printf("master_cpu = %d, master_mask = %08x, buddy_mask = %08x\n",
			  t->master_cpu, t->master_mask, t->buddy_mask);
		printf("psb_os_cpu_map = %08x, mode = %d, kseg_master = %d\n",
			  t->psb_os_cpu_map, t->mode, t->kseg_master);
		printf("app_shared_mem: addr = %016llx, size = %016llx, orig = %016llx\n",
		       (unsigned long long) t->app_shared_mem_addr,
		       (unsigned long long) t->app_shared_mem_size,
		       (unsigned long long) t->app_shared_mem_orig);
	}
}

int xlr_launch_userapp(int cpu)
{
	struct xlr_vcpu_wakeup_info *t = (struct xlr_vcpu_wakeup_info *)
		xlr_vcpu_wakeup_info + cpu;
	loading_kseg0 = 1;

#ifdef NETOS_LOADER_TODO
	if(xlr_lib_prelaunch_init_kseg) {
#else
	{
#endif /* NETOS_LOADER_TODO */
		printf("=================prelaunch_init=======================\n");
		xlr_lib_prelaunch_init_kseg(kseg_prelaunch_init_args);
	}

	update_memory_map();

	t->cpu_status = CPU_RUNNING;

	if(cpu != processor_id()){
		t->kseg_master = 1;
		loader_wakeup_cpu(cpu,t->func,t->args);
		return 0;
	}
	flush_tlb_all();
	load_scratch_info();

	dump_xlr_vcpu_wakeup_info();
	printf("Jumping to the application...\n");
	printf("------------------------------------------------------------\n");

	__asm__ volatile(
		".set push\n"
		".set noreorder\n"
		".set mips64\n"
		"move $29, %0\n"
		"move $28, %1\n"
		"nop\n"
		".set reorder\n"
		".set pop\n"
		::"r"(t->sp), "r"(t->gp)
		);

	((void (*)(int, unsigned char *[], unsigned char *[], uint64_t))t->func)
		(t->argc,t->argv,t->envs, (unsigned long)xlr_lib_boot1_info);

	/* We will never come back here
	 */
	return 0;
}

void xlr_lib_intr_handler(int irq)
{
	int cpu = processor_id();
	switch(irq){
	case IRQ_WAKEUP_CPU_IPI:
		xlr_wakeup_ipi[cpu] = 1;
		break;
	case IRQ_STOP_CPU_IPI:
		__asm__ __volatile__(
			"mtc0 %0, $14\n"
			"nop\n"
			"eret\n"
			"nop\n"
			::"r"(xlr_lib_entry)
			);
		break;
	case IRQ_RECLAIM_CPU_RSRC_IPI:
		xlr_lib_reclaim_resources();
		break;
	case IRQ_JOIN_CPU_IPI:
		printf("Rcvd JOIN IPI - CPU %d\n",cpu);
		break;
	default:
		printf("Lib Doesnt support Interupt %d\n",irq);
		break;
	}
	return;
}

extern volatile unsigned int cpu_present_map;

/* Secondary cpus come here from enable_cpus/xen_enable_cpus
 * All of them use xlr_stack_pages_temp as stack
 *
 * FIXME: what is this newmask about.
 */
void xlr_park(unsigned long args)
{
	int cpu = processor_id();

	per_cpu_init();

	disable_ie();

	init_irq();

 	ldr_trap_init();

	lib_sync();

	enable_ie();

	set_bit(cpu, &cpu_present_map);

	while (!xlr_wakeup_ipi[cpu]) {
		__asm__ volatile(
			".set noreorder\n"
			"wait\n"
			".set reorder\n"
			);
	}

	disable_ie();

	xlr_wakeup_ipi[cpu] = 0;

	enable_ie(); /* enable intr */

	xlr_start_app();
}
