#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>

#include "cpu.h"
#include "boot2.h"
#include "physaddrspace.h"
#include <string.h>
#include "system.h"
#include "memblk.h"
#include "boot_mgr.h"
#include "xlr_boot_lib.h"

#include "sys/module.h"
#include "nlm_hal_macros.h"
#include "shell.h"

int loading_nmi = 0;

extern int display_kuseg_meminfo(int argc, char *argv[]);

extern int xlr_load_kseg_app(struct bootfile_ *file, uint32_t master, int nargs, char *cmdline_args[], int loading_nmi);

void __memcpy_64(uint64_t dst, const void *src, size_t n);
void __memset_64(uint64_t dst, int c, size_t n);
void memcpy64_phys_phys(uint64_t dst, uint64_t src, size_t n);
void memcpy_virt_phys(unsigned char *dest, uint64_t src, int nbytes);
void memcpy_phys_virt(uint64_t dest, unsigned char *src, int nbytes);

void memcpy_virt_phys(unsigned char *dest, uint64_t src, int nbytes) {

    int i=0;
	
    for (i=0; i<nbytes; i++) {
        dest[i] = lb_40bit_phys_cached(src);
        src++;
    }
}

void memcpy_phys_virt(uint64_t dest, unsigned char *src, int nbytes) {

    int i=0;
    for (i=0; i<nbytes; i++) {
        sb_40bit_phys_cached(dest, (uint8_t)(src[i]));
        dest++;
    }
}

void memcpy64_phys_phys(uint64_t dst, uint64_t src, size_t n) {

    int i;
    uint64_t saddr;
    int remainder = 0;

    saddr = src;

    /* If DST and SRC are DWORD aligned */
    if (((src & 0x07) == 0) && ((dst & 0x7) == 0)) {

        int ndwords = 0;

        ndwords = n >> 3;
        remainder = n & 0x7;

        for (i = 0; i < ndwords; i+=8) {
            if (loading_nmi)
                sd_40bit_phys_uncached(dst+i, ld_40bit_phys_cached(saddr+i));
            else
                sd_40bit_phys_cached(dst+i, ld_40bit_phys_cached(saddr+i));
        }
        for (; i < ((ndwords << 3) + remainder); i+=1) {
            if (loading_nmi)
                sb_40bit_phys_uncached(dst+i, lb_40bit_phys_cached(saddr+i));
            else
                sb_40bit_phys_cached(dst+i, lb_40bit_phys_cached(saddr+i));
        }
    }
    /* If DST and SRC are WORD aligned */
    else  if (((src & 0x03) == 0) && ((dst & 0x3) == 0)) {

        int nwords = 0;

        nwords = n >> 2;
        remainder = n & 0x3;

        for (i = 0; i < nwords; i+=4) {
            if (loading_nmi)
                sw_40bit_phys_uncached(dst+i, lw_40bit_phys_cached(saddr+i));
            else
                sw_40bit_phys_cached(dst+i, lw_40bit_phys_cached(saddr+i));
        }
        for (; i < ((nwords << 2) + remainder); i+=1) {
            if (loading_nmi)
                sb_40bit_phys_uncached(dst+i, lb_40bit_phys_cached(saddr+i));
            else
                sb_40bit_phys_cached(dst+i, lb_40bit_phys_cached(saddr+i));
        }
    }
    else {
        for (i = 0; i < n; ++i) {
            if (loading_nmi)
                sb_40bit_phys_uncached(dst+i, lb_40bit_phys_cached(saddr+i));
            else
                sb_40bit_phys_cached(dst+i, lb_40bit_phys_cached(saddr+i));
        }
    }
    return;
}

void __memcpy_64(uint64_t dst, const void *src, size_t n)
{
    int i;
    char *saddr;
    int remainder = 0;

    saddr = (char *)src;

    /* If DST and SRC are DWORD aligned */
    if ((((unsigned long)src & 0x07) == 0) && (((uint64_t)dst & 0x7) == 0)) {

        int ndwords = 0;

        ndwords = n >> 3;
        remainder = n & 0x7;

        for (i = 0; i < ndwords; i+=8) {
            if (loading_nmi)
                sd_40bit_phys_uncached(dst + i, *((uint64_t *)(saddr + i)) );
            else
                sd_40bit_phys_cached(dst + i,  *((uint64_t *)(saddr + i)) );
        }
        for (; i < ((ndwords << 3) + remainder); i+=1) {
            if (loading_nmi)
                sb_40bit_phys_uncached(dst + i, (uint8_t) *(saddr + i));
            else
                sb_40bit_phys_cached(dst + i, (uint8_t) *(saddr + i));
        }
    }
    /* If DST and SRC are WORD aligned */
    else  if ((((unsigned long)src & 0x03) == 0) && (((uint64_t)dst & 0x3) == 0)) {

        int nwords = 0;
        nwords = n >> 2;
        remainder = n & 0x3;

        for (i = 0; i < nwords; i+=4) {
            if (loading_nmi)
                sw_40bit_phys_uncached(dst + i, *((uint32_t *)(saddr + i)) );
            else
                sw_40bit_phys_cached(dst + i,  *((uint32_t *)(saddr + i)) );
        }
        for (; i < ((nwords << 2) + remainder); i+=1) {
            if (loading_nmi)
                sb_40bit_phys_uncached(dst + i, (uint8_t) *(saddr + i));
            else
                sb_40bit_phys_cached(dst + i, (uint8_t) *(saddr + i));
        }
    }
    else {
        for (i = 0; i < n; ++i) {
            if (loading_nmi)
                sb_40bit_phys_uncached(dst + i, (uint8_t) *(saddr + i));
            else
                sb_40bit_phys_cached(dst + i, (uint8_t) *(saddr + i));
        }
    }

    return;
}

void __memset_64(uint64_t dst, int c, size_t n)
{
    int i = 0;
    int remainder = 0;

    /* If DST and SRC are DWORD aligned */
    if ( (((uint64_t)dst & 0x7) == 0) ) {
        int ndwords = 0;
        unsigned char val = c & 0xff;
        uint64_t value = (((uint64_t)val << 56) | 
                          ((uint64_t)val << 48) |
                          ((uint64_t)val << 40) | 
                          ((uint64_t)val << 32) |
                          ((uint64_t)val << 24) | 
                          ((uint64_t)val << 16) |
                          ((uint64_t)val << 8 ) | 
                          ((uint64_t)val << 0));

        ndwords = n >> 3;
        remainder = n & 0x7;

        for (i = 0; i < ndwords; i+=8) {
            if (loading_nmi)
                sd_40bit_phys_uncached(dst + i, value );
            else
                sd_40bit_phys_cached(dst + i, value );
        }
        for (; i < ((ndwords << 3) + remainder); i+=1) {
            if (loading_nmi)
                sb_40bit_phys_uncached(dst + i, c);
            else
                sb_40bit_phys_cached(dst + i, c);
        }
    }
    /* If DST and SRC are WORD aligned */
    else  if ( (((uint64_t)dst & 0x3) == 0) ) {
        int nwords = 0;
        unsigned char val = c & 0xff;
        uint32_t value = (((uint32_t)val << 24) | 
                          ((uint32_t)val << 16) |
                          ((uint32_t)val << 8 ) | 
                          ((uint32_t)val << 0));

        nwords = n >> 2;
        remainder = n & 0x3;

        for (i = 0; i < nwords; i+=4) {
            if (loading_nmi)
                sw_40bit_phys_uncached(dst + i, value );
            else
                sw_40bit_phys_cached(dst + i, value );
        }
        for (; i < ((nwords << 2) + remainder); i+=1) {
            if (loading_nmi)
                sb_40bit_phys_uncached(dst + i, c);
            else
                sb_40bit_phys_cached(dst + i, c);
        }
    }
    else {
        for (i = 0; i < n; ++i) {
            if (loading_nmi)
                sb_40bit_phys_uncached(dst + i, (uint8_t)c);
            else
                sb_40bit_phys_cached(dst + i, (uint8_t) c);
        }
    }
    return;
}

static char *elfload_usage = 
"----------------------------------------------------------\n"
"elfload <options>\n"
"   -a     physical address to elfload from\n"
"   -h     help\n"
"----------------------------------------------------------\n";

/*
 * FIXME: remove the 'K' option in the following
 *        as it is no longer needed to load mapped
 *        kernels, thanks to linker script magic
 */

int elfload_shell_cmd(int argc, char *argv[]) {

    uint64_t start;
    int ch, res;
    unsigned char *sStart = NULL;
    struct bootfile_ *this_file = get_bootfile();

    if(this_file == NULL)
        return -1;

    res = start = 0;

    phnx_initialize_getopt();
    while ((ch = getopt(argc, argv, "a:KhA:T:")) != -1) {
        switch(ch) {
			
        case 'a':
			sStart = (unsigned char *)optarg;
			break;
		case 'h':
			printf("%s", elfload_usage);
			return 0;
		case 'A':
		case 'T':
		case 'K':
		default:
			break;
        }
    }
    
    if(sStart) {
		start = strtoull((const char *)sStart,0,16);
        this_file->start = start;
    }
    else {
        /* see if a tftp-ed file is available */
        if (!this_file->valid) {
            printf("No image found to load!\n");
            return -1;
        }
        else {
            start = this_file->start;
        }
    }

    res = xlr_load_kseg_app(this_file, 0x0, argc, argv, loading_nmi);

    this_file->busy     = 0;
    this_file->valid    = 0;
    this_file->loaded   = 1;

    return res;
}

void elf_initialize(void)
{
    xlr_load_init();

    register_shell_cmd("elfload", (void *)elfload_shell_cmd, "validate elf file and load into mem");
    register_shell_cmd("kuseg_mem",(void *)display_kuseg_meminfo, "display available KUSEG mem");

    display_kuseg_meminfo(0, 0);
}
